# SPDX-License-Identifier: Apache-2.0

"""
This example show how to infer or call a model using the framework
and an Ollama backend to serve the model.
In this scenario the chat request enables thinking mode in the model
to provide better understanding of how the model arrived at its answer.
"""

# Local
from granite_io import make_backend, make_io_processor
from granite_io.types import ChatCompletionInputs, UserMessage

model_name = "granite3.2:8b"
io_processor = make_io_processor(
    model_name, backend=make_backend("openai", {"model_name": model_name})
)
question = "Find the fastest way for a seller to visit all the cities in their region"
messages = [UserMessage(content=question)]

# With Thinking
outputs = io_processor.create_chat_completion(
    ChatCompletionInputs(messages=messages, thinking=True)
)
print("------ WITH THINKING ------")
print(">> Thoughts:")
print(outputs.results[0].next_message.reasoning_content)
print(">> Response:")
print(outputs.results[0].next_message.content)



# SPDX-License-Identifier: Apache-2.0

# Standard
import os

# Third Party
from test_utils import fix_granite_date, load_text_file

# Local
from granite_io import get_input_processor
from granite_io.types import (
    ChatCompletionInputs,
    UserMessage,
)

_GENERALE_MODEL_NAME = "Granite 3.3"
_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), "testdata")


def test_run_processor_reasoning():
    input_processor = get_input_processor(_GENERALE_MODEL_NAME)
    question = (
        "Find the fastest way for a seller to visit all the cities in their region"
    )
    messages = [UserMessage(content=question)]
    prompt = input_processor.transform(
        ChatCompletionInputs(messages=messages, thinking=True)
    )

    expected_prompt = load_text_file(
        os.path.join(_TEST_DATA_DIR, "test_reasoning_prompt.txt")
    )
    assert isinstance(prompt, str)
    assert prompt == fix_granite_date(expected_prompt)



# SPDX-License-Identifier: Apache-2.0

"""
This example show how to infer or call a model using the framework
and an Ollama backend to serve the model. It uses the framework capability
to create a custom IO processor using a specific input processor and output
processor.

In this scenario the chat request enables thinking mode in the model
to provide better understanding of how the model arrived at its answer.
"""

# Third Party
import aconfig

# Local
from granite_io import (  # make_new_io_processor,
    Backend,
    ChatCompletionInputs,
    ChatCompletionResults,
    GenerateResults,
    ModelDirectInputOutputProcessor,
    UserMessage,
    get_input_processor,
    get_output_processor,
    make_backend,
)

ollama_model_name = "granite3.2:8b"

# Create a custom IO processor using specified input and output processors
# This use case uses the 'make_new_io_processor' wrapper function.
# io_processor = make_new_io_processor(
#    input_processor=get_input_processor(ollama_model_name),
#    output_processor=get_output_processor(ollama_model_name),
#    backend=make_backend("openai", {"model_name": ollama_model_name}),
# )


# Alternative approach to creating a custom IO processor. In this scenario
# we will inherit from base class 'ModelDirectInputOutputProcessor' and add
# custom input and output processing to it. In this case will use input and output
# processors from the library. You can also implement without using input and output
# processors.
class _MyInputOutputProcessor(ModelDirectInputOutputProcessor):
    """
    Custom IO processor which uses specifc Granite 3.2 input and output
    processors.
    """

    def __init__(
        self,
        model_name: str,
        config: aconfig.Config = None,
        backend: Backend | None = None,
    ):
        """
        :param backend: Handle on inference engine, required if this io processor's
            :func:`create_chat_completion()` method is going to be used
        :param config: Setup config for this IO processor
        """
        super().__init__(backend=backend)
        self._model_name = model_name

    def inputs_to_string(
        self, inputs: ChatCompletionInputs, add_generation_prompt: bool = True
    ) -> str:
        input_processor = get_input_processor(self._model_name)
        return input_processor.transform(inputs, add_generation_prompt)

    def output_to_result(
        self,
        output: GenerateResults,
        inputs: ChatCompletionInputs | None = None,
    ) -> ChatCompletionResults:
        output_processor = get_output_processor(self._model_name)
        return output_processor.transform(output, inputs)


# Create an instance of the custom IO processor using input and outut processors
# identified/specific by model name
io_processor = _MyInputOutputProcessor(
    model_name=ollama_model_name,
    backend=make_backend("openai", {"model_name": ollama_model_name}),
)

question = "Find the fastest way for a seller to visit all the cities in their region"
messages = [UserMessage(content=question)]

# With Thinking
outputs = io_processor.create_chat_completion(
    ChatCompletionInputs(messages=messages, thinking=True)
)
print("------ WITH THINKING ------")
print(">> Thoughts:")
print(outputs.results[0].next_message.reasoning_content)
print(">> Response:")
print(outputs.results[0].next_message.content)



# SPDX-License-Identifier: Apache-2.0

"""
The io module holds the abstraction and implementations for model-specific IO
processing.
"""

# Local
from granite_io.io.base import InputOutputProcessor  # noqa: F401
from granite_io.io.base import InputProcessor  # noqa: F401
from granite_io.io.base import OutputProcessor  # noqa: F401
from granite_io.io.registry import (  # noqa: F401
    get_input_processor,
    get_output_processor,
    input_processor,
    io_processor,
    make_io_processor,
    output_processor,
)
import granite_io.io.granite_3_2.granite_3_2  # noqa: F401
import granite_io.io.granite_3_3.granite_3_3  # noqa: F401



# SPDX-License-Identifier: Apache-2.0

"""
Utilities for optional dependencies
"""

# Standard
from contextlib import contextmanager
import logging


@contextmanager
def import_optional(extra_name: str):
    """Context manager to handle optional imports"""
    try:
        yield
    except ImportError as err:
        logging.warning(
            "%s.\nHINT: You may need to pip install %s[%s]",
            err,
            __package__,
            extra_name,
        )
        raise



# SPDX-License-Identifier: Apache-2.0

"""
Test cases for io_adapters/answerability.py
"""

# Standard
import datetime
import textwrap

# Third Party
import pytest

# Local
from granite_io.backend.vllm_server import LocalVLLMServer
from granite_io.io.answerability import AnswerabilityIOProcessor
from granite_io.io.granite_3_2.input_processors.granite_3_2_input_processor import (
    Granite3Point2Inputs,
    override_date_for_testing,
)
from granite_io.types import GenerateResult, GenerateResults

_EXAMPLE_CHAT_INPUT = Granite3Point2Inputs.model_validate(
    {
        "messages": [
            {"role": "assistant", "content": "Welcome to pet questions!"},
            {"role": "user", "content": "Which of my pets have fleas?"},
        ],
        "documents": [
            {"text": "My dog has fleas."},
            {"text": "My cat does not have fleas."},
        ],
        "generate_inputs": {
            "temperature": 0.0  # Ensure consistency across runs
        },
    }
)


def _make_result(content: str):
    """Convenience method to create a fake model output object."""
    return GenerateResult(
        completion_string=content, completion_tokens=[], stop_reason="dummy stop reason"
    )


_TODAYS_DATE = datetime.datetime.now().strftime("%B %d, %Y")


def test_canned_input():
    """
    Validate that the I/O processor handles a single instance of canned input in the
    expected way.
    """
    io_processor = AnswerabilityIOProcessor(None)
    output = io_processor.inputs_to_generate_inputs(_EXAMPLE_CHAT_INPUT).prompt
    print("*****")
    print(output)
    print("*****")
    expected_output = textwrap.dedent(f"""\
    <|start_of_role|>system<|end_of_role|>Knowledge Cutoff Date: April 2024.
    Today's Date: {_TODAYS_DATE}.
    You are Granite, developed by IBM.Write the response to the user's input by \
strictly aligning with the facts in the provided documents. If the information needed \
to answer the question is not available in the documents, inform the user that the \
question cannot be answered based on the available data.<|end_of_text|>
    <|start_of_role|>documents<|end_of_role|>Document 0
    My dog has fleas.
    
    Document 1
    My cat does not have fleas.<|end_of_text|>
    <|start_of_role|>assistant<|end_of_role|>Welcome to pet questions!<|end_of_text|>
    <|start_of_role|>user<|end_of_role|>Which of my pets have fleas?<|end_of_text|>
    <|start_of_role|>answerability<|end_of_role|>""")
    assert output == expected_output


def test_canned_output():
    """
    Validate that the I/O processor handles a single instance of canned model output
    in the expected way.
    """
    io_processor = AnswerabilityIOProcessor(None)

    raw_output_to_expected = [
        ("answerable", "answerable"),
        ("unanswerable", "unanswerable"),
        ("<invalid model response>", "ERROR"),
    ]

    # Single output
    for raw_output, expected in raw_output_to_expected:
        output = io_processor.output_to_result(
            GenerateResults(results=[_make_result(raw_output)]), _EXAMPLE_CHAT_INPUT
        )
        assert len(output.results) == 1
        assert output.results[0].next_message.content == expected

    # Multiple outputs
    multi_raw_output = [
        _make_result(raw_output) for raw_output, _ in raw_output_to_expected
    ]
    multi_expected = [expected for _, expected in raw_output_to_expected]
    multi_output = io_processor.output_to_result(
        GenerateResults(results=multi_raw_output), _EXAMPLE_CHAT_INPUT
    )
    multi_output_strs = [r.next_message.content for r in multi_output.results]
    assert multi_output_strs == multi_expected


@pytest.mark.vcr
def test_run_model(lora_server: LocalVLLMServer, fake_date: str):
    """
    Run a chat completion through the LoRA adapter using the I/O processor.
    """
    backend = lora_server.make_lora_backend("answerability")
    io_proc = AnswerabilityIOProcessor(backend)

    # Pass our example input thorugh the I/O processor and retrieve the result
    override_date_for_testing(fake_date)  # For consistent VCR output
    chat_result = io_proc.create_chat_completion(_EXAMPLE_CHAT_INPUT)

    assert chat_result.results[0].next_message.content in ("answerable", "unanswerable")



# SPDX-License-Identifier: Apache-2.0


# Standard
from collections.abc import Callable

# Local
from granite_io.io.base import ChatCompletionResults, InputOutputProcessor
from granite_io.optional import import_optional
from granite_io.types import ChatCompletionInputs


def integer_normalizer(model_outputs: list[str]) -> list[str]:
    """
    Simple normalizer for integer answers. Parses each input string to an int and
    returns the normalized number. Non-integral values turn into the string "nan".
    """
    results = []
    for model_output in model_outputs:
        try:
            results.append(str(int(model_output)))
        except ValueError:
            results.append("nan")
    return results


class MajorityVotingProcessor(InputOutputProcessor):
    """
    Implementation of basic single-step majority sampling.
    """

    _generator: InputOutputProcessor
    _normalizer: Callable
    _samples_per_completion: int

    def __init__(
        self,
        generator: InputOutputProcessor,
        normalizer: Callable[[list[str]], list[str]],
        samples_per_completion: int = 100,
    ):
        """
        :param generator: Sub-processor over which this processor should sample
        :param normalizer: Function that takes multiple results and returns normalized
            versions of said results for grouping
        :param samples_per_completion: Number of samples to take per desired completion
        """
        self._generator = generator
        self._normalizer = normalizer
        self._samples_per_completion = samples_per_completion

    async def acreate_chat_completion(
        self, inputs: ChatCompletionInputs
    ) -> ChatCompletionResults:
        with import_optional("pandas"):
            # Third Party
            import pandas as pd

        # Step 1: Collect samples from the "generator" sub-processor
        original_n = (
            1
            if inputs.generate_inputs is None or inputs.generate_inputs.n is None
            else inputs.generate_inputs.n
        )
        num_samples = self._samples_per_completion * original_n
        base_results = await self._generator.acreate_chat_completion(
            inputs.with_addl_generate_params({"n": num_samples})
        )

        # Step 2: Normalize the samples
        sample_contents = [s.next_message.content for s in base_results.results]
        normalized_contents = self._normalizer(sample_contents)
        assert len(sample_contents) == len(normalized_contents)  # Sanity check

        # Step 3: Group and aggregate
        indices = (
            # Group results by normalized contents, sort the groups by count, and use
            # the first result in each group as the representative of the group
            pd.DataFrame(
                {
                    "result_num": range(len(sample_contents)),
                    "normalized_result": normalized_contents,
                }
            )
            .groupby("normalized_result")
            .agg({"normalized_result": "count", "result_num": "min"})
            .rename(columns={"normalized_result": "count"})
            .sort_values(["count", "result_num"], ascending=False)
        )["result_num"].to_list()

        # If the caller requested multiple completions, return multiple results that
        # each have different normalized forms.
        num_outputs = min(len(indices), original_n)

        return ChatCompletionResults(
            results=[base_results.results[indices[i]] for i in range(num_outputs)]
        )



# SPDX-License-Identifier: Apache-2.0


# Local
from .citations import CitationsCompositeIOProcessor, CitationsIOProcessor

# Expose public symbols at `granite_io.io.certainty` to save users from typing
__all__ = ["CitationsIOProcessor", "CitationsCompositeIOProcessor"]



# SPDX-License-Identifier: Apache-2.0

# Standard
from typing import TYPE_CHECKING, Any
import asyncio
import concurrent.futures
import dataclasses

# Third Party
import aconfig

# Local
from granite_io.backend.base import Backend
from granite_io.backend.registry import backend
from granite_io.optional import import_optional
from granite_io.types import GenerateInputs, GenerateResult, GenerateResults

if TYPE_CHECKING:
    # Third Party
    import transformers


def _detect_hw_accel():
    """:returns: Pytorch's name for the best available hardware accelerator on the
    current machine."""
    with import_optional("transformers"):
        # Third Party
        import torch
    if torch.cuda.is_available():
        result = "cuda"
    elif torch.backends.mps.is_available():
        result = "mps"
    else:
        result = "cpu"
    return result


@dataclasses.dataclass
class _GenerationInputs:
    """Dataclass to encapsulate inputs for calling the Transformers generate() method"""

    generation_config: "transformers.GenerationConfig"

    # Re-enable to use constrained decoding
    # prefix_allowed_tokens_fn: Optional[Callable[[int, torch.Tensor], List[int]]]
    model_input: dict


@backend(
    "transformers",
    config_schema={
        "properties": {
            "model_name": {"type": "string"},
            "device": {"type": "string"},
        }
    },
)
class TransformersBackend(Backend):
    _model_str: str
    _torch_device_name: str
    "Device used for hardware acceleration by this backend."
    _model: "transformers.AutoModelForCausalLM"
    _tokenizer: "transformers.AutoTokenizer"
    _executor: concurrent.futures.ThreadPoolExecutor
    "Single background thread, wrapped in an executor for queueing"

    def __init__(self, config: aconfig.Config):
        """
        :param model_to_load: String that the Transformers library can map to a location
            from which to load model files. This is usually a set of HuggingFace model
            coordinates such as "ibm-granite/granite-3.0-8b-instruct", but it can also
            be a local filesystem path or a URL.
        :param device: Optional Pytorch device string, usually "cpu" or "cuda"
        """
        super().__init__(config)

        with import_optional("transformers"):
            # Third Party
            import torch  # noqa: F401
            import transformers

        self._torch_device_name = config.device if config.device else _detect_hw_accel()
        if self._torch_device_name == "cpu":
            # CPU mode; prevent thrashing
            torch.set_num_threads(4)
        self._model = transformers.AutoModelForCausalLM.from_pretrained(
            config.model_name
        ).to(self._torch_device_name)
        self._tokenizer = transformers.AutoTokenizer.from_pretrained(config.model_name)
        self._executor = concurrent.futures.ThreadPoolExecutor()

    async def pipeline(self, inputs: GenerateInputs) -> GenerateResults:
        """
        Process input, call completion (generate), process and return output

        Override to pass inputs for prefix tokens trimming.
        """
        inputs = self.process_input(inputs)
        outputs = await self.generate(inputs)
        return self.process_output(outputs=outputs, inputs=inputs)

    async def generate(self, inputs: _GenerationInputs) -> GenerateResults:
        # Wrap the call to the non-async Transformers library in a thread pool
        concurrent_futures_future = self._executor.submit(
            self._generate_callback,
            inputs,
        )
        async_future = asyncio.wrap_future(concurrent_futures_future)
        return await async_future

    def process_input(self, inputs: GenerateInputs) -> GenerateInputs:
        """Subroutine that encapsulates all the prerequisites
        that are necessary to call ``AutoModelForCausalLM.generate()``."""

        kwargs = super().process_input(inputs).model_dump(exclude_unset=True)

        # Migrate alias kwargs to this flavor of backend
        self.kwarg_alias(kwargs, "stop_strings", "stop")
        self.kwarg_alias(kwargs, "num_return_sequences", "n")

        # num_beams > num_return_sequences
        n = kwargs.get("num_return_sequences", 1)
        if n is not None and n < 1:
            raise ValueError(f"Invalid value for num_return_sequences ({n})")
        if n is not None:
            num_beams = kwargs.get("num_beams", 1)
            if num_beams is None or num_beams < n:
                num_beams = n

        # Import packages from extras "transformers"
        with import_optional("transformers"):
            # Third Party
            import torch  # noqa: F401
            import transformers

        # Turn the conversation prefix into tokens for input to the model.
        model_input = self._tokenizer(
            # The conversation up to this point
            kwargs.get("prompt"),
            # Tell the tokenizer to return a tensor instead of a Python list.
            # The model expects to receive tensors as input, so you almost always
            # need to set this.
            # You must manually select a framework. Good luck switching frameworks
            # afterwards. Here we use PyTorch.
            # Enabling tensor output also CHANGES THE SHAPE OF THE OUTPUT from
            # a 1D list to a 2D singleton batch. The model can only consume batches.
            # The tokenizer's decode() method can only consume 1D lists.
            # This tensor will be created on the framework's current default
            # device, so be sure to set that default appropriately.
            return_tensors="pt",
        )

        # AutoTokenizer uses two different tokenizer classes internally. One
        # of these classes has the ability to put tensors on the device they
        # should be on, while the other does not. Since we can't predict which
        # implementation we'll have, we need to assume that everything's on
        # the wrong device and move it to the right device. This of course
        # requires transforming the values under the keys of a dictionary.
        model_input = {
            k: v.to(self._torch_device_name) if isinstance(v, torch.Tensor) else v
            for k, v in model_input.items()
        }

        # transformers needs the tokenizer when stop_strings are given
        if kwargs.get("stop_strings"):
            model_input["tokenizer"] = self._tokenizer

        # The generate() method sometimes needs to know what is the integer ID
        # of the padding token, and for some reason this critical piece of information
        # isn't included in the serialized model. We get it from the tokenizer.
        # And of course some tokenizers don't set this parameter, in which case
        # we use the end of string token and hope for the best.
        pad_token_id = self._tokenizer.pad_token_id
        if pad_token_id is None:
            pad_token_id = self._tokenizer.eos_token_id
        if pad_token_id is None:
            # Raise an error here because the some branches of the generate
            # method won't complain about an invalid value of this parameter,
            # while others will raise a cryptic exception from deep within
            # their beam search code.
            raise ValueError(
                f"Couldn't figure out padding token for tokenizer {self._tokenizer}"
            )

        # The supported way to pass parameters to the generate() method is
        # to pass them to the constructor for another class, then pass the
        # resulting object to model.generate().
        generation_config = transformers.GenerationConfig(
            max_new_tokens=kwargs.get("max_new_tokens", 1024),  # TEMPORARY
            num_return_sequences=n,
            num_beams=num_beams,
            # Scores are wrong anyhow, so don't bother
            output_scores=False,
            # If you don't set this flag, you'll get a string back instead of
            # a collection of multiple tensors and lists.
            return_dict_in_generate=True,
            # VERY important parameter, with no documentation on what's the right value.
            # The right value varies by model and by application.
            # Wrong values (often including the default) will produce very bad output.
            # LOWER values result in MORE penalty for repetition, because of course they
            # do.
            repetition_penalty=kwargs.get("repetition_penalty", 1.0),
            top_p=kwargs.get("top_p", 1.0),
            top_k=kwargs.get("top_k", 50),
            temperature=kwargs.get("temperature", 1.0),
            # See long note above.
            pad_token_id=pad_token_id,
            # Make sure you specify this token explicitly, or you will have
            # a bad time.
            eos_token_id=self._tokenizer.eos_token_id,
            stop_strings=kwargs.get("stop_strings"),
        )

        # Parameters for constrained generation are **not** passed to generate()
        # via the GenerationConfig object, but are instead passed in as a
        # separate argument to the generate() method, containing a callback
        # that itself must contain a pointer to the model's tokenizer.

        # Re-enable to use constrained decoding
        # prefix_allowed_tokens_fn=_response_format_to_transformers_callback(
        #     tokenizer, response_format
        # )

        return _GenerationInputs(
            generation_config=generation_config,
            model_input=model_input,
            # prefix_allowed_tokens_fn=prefix_allowed_tokens_fn
        )

    def process_output(self, outputs, *, inputs=None):
        # The result of generate() is of course the prompt concatenated with the
        # additional tokens generated. Strip off the prompt.
        generated_results = []

        # Generation inputs are needed for special processing below
        input_len = len(inputs.model_input["input_ids"][0]) if inputs else 0

        # for i, sequence in enumerate(model_output.sequences):
        for sequence in outputs.sequences:
            full_token_sequence = sequence.cpu().tolist()
            generated_tokens = full_token_sequence[input_len:]

            # The generate() method doesn't explicitly tell us why it stopped
            # generating. We are supposed to infer that from the output.
            if generated_tokens[-1] == self._tokenizer.eos_token_id:
                stop_reason = "end_of_turn"
                # We're also supposed to strip off the end-of-turn tokens ourselves.
                generated_tokens = generated_tokens[:-1]

                # When one requests multiple completions, the shorter completions will
                # come out padded with extra end-of-turn tokens so that everything is
                # the same length.
                while (
                    len(generated_tokens) > 0
                    and generated_tokens[-1] == self._tokenizer.eos_token_id
                ):
                    generated_tokens = generated_tokens[:-1]

            else:
                stop_reason = "out_of_tokens"

            # Of course, the model does not have a pointer to its tokenizer, so
            # we need to post-process the model's output to get a usable string.
            completion_string = self._tokenizer.decode(generated_tokens)

            # Now we can see if the stop reason was actually a stop string.
            if stop_strings := inputs.generation_config.stop_strings:
                if isinstance(stop_strings, str):
                    stop_strings = [stop_strings]
                for s in stop_strings:
                    if completion_string.endswith(s):
                        completion_string = completion_string.removesuffix(s)
                        stop_reason = "stop"
                        break

            generated_results.append(
                GenerateResult(
                    completion_string=completion_string,
                    completion_tokens=generated_tokens,
                    stop_reason=stop_reason,
                )
            )

        return GenerateResults(results=generated_results)

    def _generate_callback(self, generation_inputs: _GenerationInputs) -> Any:
        """
        :param generation_inputs: Wrapper for the required inputs to a Transformers
            :func:`generate()` call, prepared in the event loop thread.

        :returns: Whatever type this model's `generate()` method happens to return on
            this day of the week, which changes without warning from one version to the
            next of the library.
        """
        with import_optional("transformers"):
            # Third Party
            import torch

        def do_generate():
            return self._model.generate(
                **(generation_inputs.model_input),
                generation_config=generation_inputs.generation_config,
            )

        if self._torch_device_name == "cuda":
            # Make sure computations for this thread will happen in a separate CUDA
            # context.
            stream = torch.cuda.Stream()
            with torch.cuda.stream(stream):
                return do_generate()
        else:
            return do_generate()



# SPDX-License-Identifier: Apache-2.0


# Local
from .hallucinations import (
    HallucinationsCompositeIOProcessor,
    HallucinationsIOProcessor,
)

# Expose public symbols at `granite_io.io.hallucinations` to save users from typing
__all__ = [
    "HallucinationsIOProcessor",
    "HallucinationsCompositeIOProcessor",
]



# SPDX-License-Identifier: Apache-2.0

# Standard
import os

# Third Party
from test_utils import load_text_file

# Local
from granite_io import get_output_processor
from granite_io.types import (
    ChatCompletionInputs,
    GenerateResult,
    GenerateResults,
    UserMessage,
)

_GENERAL_MODEL_NAME = "Granite 3.3"
_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), "testdata")


def test_run_processor_reasoning():
    output_processor = get_output_processor(_GENERAL_MODEL_NAME)
    # Process the output
    results = []
    raw_output = load_text_file(
        os.path.join(_TEST_DATA_DIR, "test_raw_reasoning_output.txt")
    )
    results.append(
        GenerateResult(
            completion_string=raw_output,
            completion_tokens=[],
            stop_reason="stop",
        )
    )
    question = (
        "Find the fastest way for a seller to visit all the cities in their region"
    )
    messages = [UserMessage(content=question)]
    outputs = output_processor.transform(
        GenerateResults(results=results),
        ChatCompletionInputs(messages=messages, thinking=True),
    )

    expected_thought_output = load_text_file(
        os.path.join(_TEST_DATA_DIR, "test_reasoning_output_processor_thought.txt")
    )
    assert expected_thought_output == outputs.results[0].next_message.reasoning_content

    expected_response_output = load_text_file(
        os.path.join(_TEST_DATA_DIR, "test_reasoning_output_processor_response.txt")
    )
    assert expected_response_output == outputs.results[0].next_message.content


def test_parse_tool_output():
    # Example from the Granite 3.3 documentation.
    # And yes, this output is wrong. The "location" argument is supposed to be
    # "New York, NY". But this is what the model returns when given this input.
    TOOL_REQUEST = ChatCompletionInputs(
        messages=[
            {"role": "user", "content": "What's the current weather in New York?"}
        ],
        tools=[
            {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "arguments": {
                    "location": {
                        "description": "The city and state, e.g. San Francisco, CA"
                    }
                },
            },
            {
                "name": "get_stock_price",
                "description": "Retrieves the current stock price for a given ticker "
                "symbol. The ticker symbol must be a valid symbol for a publicly "
                "traded company on a major US stock exchange like NYSE or NASDAQ. "
                "The tool will return the latest trade price in USD. It should be "
                "used when the user asks about the current or most recent price of "
                "a specific stock. It will not provide any other information about "
                "the stock or company.",
                "arguments": {
                    "ticker": {
                        "description": "The stock ticker symbol, e.g. AAPL for Apple "
                        "Inc."
                    }
                },
            },
        ],
    )
    TOOL_OUTPUT = (
        '<tool_call>[{"name": "get_current_weather", "arguments": '
        '{"location": "New York"}}]'
    )

    output_processor = get_output_processor(_GENERAL_MODEL_NAME)
    outputs = output_processor.transform(
        GenerateResults(
            results=[
                GenerateResult(
                    completion_string=TOOL_OUTPUT,
                    completion_tokens=[],
                    stop_reason="stop",
                )
            ]
        ),
        TOOL_REQUEST,
    )

    # Can't just check the result in one step due to random function IDs
    assert len(outputs.results) == 1
    assert len(outputs.results[0].next_message.tool_calls) == 1
    assert outputs.results[0].next_message.tool_calls[0].name == "get_current_weather"
    assert outputs.results[0].next_message.tool_calls[0].arguments == {
        "location": "New York"
    }



# SPDX-License-Identifier: Apache-2.0

# Standard
import collections.abc

# Third Party
import pytest
import torch

# Local
from granite_io import make_backend
from granite_io.backend import Backend
from granite_io.backend.vllm_server import LocalVLLMServer
from granite_io.io.consts import (
    _GRANITE_3_2_2B_HF,
    _GRANITE_3_3_2B_HF,
    _GRANITE_3_3_2B_OLLAMA,
)
from granite_io.io.granite_3_2.input_processors.granite_3_2_input_processor import (
    override_date_for_testing as g32_override_date_for_testing,
)


@pytest.fixture(scope="session")
def vcr_config():
    return {"filter_headers": ["authorization"]}


def backend_openai() -> Backend:
    return make_backend(
        "openai",
        {
            "model_name": "granite3.2:2b",
            "openai_api_key": "ollama",
            "openai_base_url": "http://localhost:11434/v1",
        },
    )


def backend_litellm() -> Backend:
    return make_backend(
        "litellm",
        {
            "model_name": "ollama/" + "granite3.2:2b",
        },
    )


def backend_transformers() -> Backend:
    return make_backend(
        "transformers",
        {
            "model_name": _GRANITE_3_2_2B_HF,
        },
    )


def backend_3_3_openai() -> Backend:
    return make_backend(
        "openai",
        {
            "model_name": _GRANITE_3_3_2B_OLLAMA,
            "openai_api_key": "ollama",
            "openai_base_url": "http://localhost:11434/v1",
        },
    )


def backend_3_3_litellm() -> Backend:
    return make_backend(
        "litellm",
        {
            "model_name": "ollama/" + _GRANITE_3_3_2B_OLLAMA,
        },
    )


def backend_3_3_transformers() -> Backend:
    return make_backend(
        "transformers",
        {
            "model_name": _GRANITE_3_3_2B_HF,
        },
    )


@pytest.fixture(scope="function")
def fake_date():
    """
    We need today's date to be constant so that ``vcrpy`` functions on our prompts.

    By wrapping the creation of this date in a fixture, we can be sure to reset to
    normal behavior if a test fails.

    :returns: a fake version of today's date, for use in prompt text that mentions the
     date.
    """
    yield "April 1, 2025"

    # Cleanup code. Augment as needed as we add new IO processors with date-dependent
    # prompts.
    g32_override_date_for_testing(None)


@pytest.fixture(
    scope="session", params=[backend_openai, backend_litellm, backend_transformers]
)
def backend_x(request) -> Backend:
    return request.param()


@pytest.fixture(
    scope="session",
    params=[backend_3_3_openai, backend_3_3_litellm, backend_3_3_transformers],
)
def backend_3_3(request) -> Backend:
    return request.param()


@pytest.fixture(scope="session")
def lora_server() -> collections.abc.Generator[LocalVLLMServer, object, None]:
    """
    Fixture that runs a local vLLM server. The server uses a fixed port because the
    ``vcrpy`` package requires fixed local ports.

    :returns: vLLM server with all the LoRAs for which we currently have IO processors

    """
    if not torch.cuda.is_available():
        pytest.xfail("GPU required to run vLLM. vLLM required to run model.")

    # Currently all adapters are trained on Granite 3.2 8B
    base_model = "ibm-granite/granite-3.2-8b-instruct"
    lora_adapters = [
        # Format is server name, model name
        (
            "answerability",
            "ibm-granite/granite-3.2-8b-lora-rag-answerability-prediction",
        ),
        ("certainty", "ibm-granite/granite-uncertainty-3.2-8b-lora"),
        ("citations", "ibm-granite/granite-3.2-8b-lora-rag-citation-generation"),
        (
            "hallucinations",
            "ibm-granite/granite-3.2-8b-lora-rag-hallucination-detection",
        ),
        ("query_rewrite", "ibm-granite/granite-3.2-8b-lora-rag-query-rewrite"),
    ]

    server = LocalVLLMServer(base_model, lora_adapters=lora_adapters, port=35782)
    server.wait_for_startup(200)
    yield server

    # Shutdown code runs at end of test session
    server.shutdown()



# SPDX-License-Identifier: Apache-2.0


# Local
from .certainty import (
    AssistantMessageWithScore,
    CertaintyCompositeIOProcessor,
    CertaintyIOProcessor,
)

# Expose public symbols at `granite_io.io.certainty` to save users from typing
__all__ = [
    "CertaintyIOProcessor",
    "CertaintyCompositeIOProcessor",
    "AssistantMessageWithScore",
]



# SPDX-License-Identifier: Apache-2.0

"""
This example show how to process input to a model(prompt) and putput from the model
using the framework.

The inference of the model is managed independent of the framework using Ollama backend
to serve the model.

In this scenario the chat request enables thinking mode in the model to provide better
understanding of how the model arrived at its answer.
"""

# Third Party
import openai

# Local
from granite_io import get_input_processor, get_output_processor
from granite_io.types import (
    ChatCompletionInputs,
    GenerateResult,
    GenerateResults,
    UserMessage,
)

api_key = "ollama"
base_url = "http://localhost:11434/v1"  # Ollama server hosting model
ollama_model_name = "granite3.2:8b"

# Connect to Ollama server
openai_client = openai.OpenAI(base_url=base_url, api_key=api_key)

# Get input processor for the model type and version
input_processor = get_input_processor(
    ollama_model_name,
)

# Construct the prompt for Granite 3.2 model so that it will provide reasoning or
# thinking of how it provided its answer to the question asked. It will be similar
# to this format:
#
# prompt = """<|start_of_role|>system<|end_of_role|>Knowledge Cutoff Date: April 2024.
# Today's Date: March 13, 2025.
# You are Granite, developed by IBM. You are a helpful AI assistant.
# Respond to every user query in a comprehensive and detailed way. You can write down \
# your thoughts and reasoning process before responding. In the thought process, \
# engage in a comprehensive cycle of analysis, summarization, exploration, \
# reassessment, reflection, backtracing, and iteration to develop well-considered \
# thinking process. In the response section, based on various attempts, explorations, \
# and reflections from the thoughts section, systematically present the final \
# solution that you deem correct. The response should summarize the thought process. \
# Write your thoughts after 'Here is my thought process:' and write your response \
# after 'Here is my response:' for each user query.<|end_of_text|>
# <|start_of_role|>user<|end_of_role|>Find the fastest way for a seller to visit all \
# the cities in their region<|end_of_text|>
# <|start_of_role|>assistant<|end_of_role|>"""
question = "Find the fastest way for a seller to visit all the cities in their region"
messages = [UserMessage(content=question)]
prompt = input_processor.transform(
    ChatCompletionInputs(messages=messages, thinking=True)
)

# Infer/query the model
result = openai_client.completions.create(
    model=ollama_model_name,
    prompt=prompt,
)

# Process the output as standard
print("\n\n *** Standard Output ***\n\n")
for choice in result.choices:
    print(f"{choice.text}\n")

# Process output using granite-io output processor. Output processor parses the output
# into response and reasoning for you

# Get output processor for the model type and version
output_processor = get_output_processor(
    ollama_model_name,
)

# Process the output
results = []
for choice in result.choices:
    results.append(
        GenerateResult(
            completion_string=choice.text,
            completion_tokens=[],  # Not part of the OpenAI spec
            stop_reason=choice.finish_reason,
        )
    )
outputs = output_processor.transform(
    GenerateResults(results=results),
    ChatCompletionInputs(messages=messages, thinking=True),
)

print("\n\n *** granite-io Output ***\n\n")
print(">> Thoughts:")
print(outputs.results[0].next_message.reasoning_content)
print("\n>> Response:")
print(outputs.results[0].next_message.content)



# SPDX-License-Identifier: Apache-2.0

# Standard
from typing import TYPE_CHECKING

# Third Party
import aconfig

# Local
from granite_io.backend.base import Backend
from granite_io.backend.registry import backend
from granite_io.optional import import_optional
from granite_io.types import GenerateInputs, GenerateResult, GenerateResults

if TYPE_CHECKING:
    # Third Party
    import openai


@backend(
    "openai",
    config_schema={
        "properties": {
            "model_name": {"type": "string"},
            "openai_base_url": {"type": "string"},
            "openai_api_key": {"type": "string"},
        }
    },
)
class OpenAIBackend(Backend):
    _openai_client: "openai.AsyncOpenAI"

    def __init__(self, config: aconfig.Config):
        super().__init__(config)
        with import_optional("openai"):
            # Third Party
            import openai

        api_key = config.get("openai_api_key", "ollama")
        base_url = config.get("openai_base_url", "http://localhost:11434/v1")
        default_headers = config.get("default_headers")

        self._openai_client = openai.AsyncOpenAI(
            base_url=base_url, api_key=api_key, default_headers=default_headers
        )

    async def generate(self, inputs: GenerateInputs):
        """Run a direct /completions call"""

        # Implementations of the OpenAI APIs tend to get upset if you pass them null
        # values, particularly if those values are associated with arguments that the
        # implementation doesn't support. So strip out null parameters.
        args_dict = inputs.model_dump(exclude_none=True)
        # pylint: disable-next=missing-kwoa
        return await self._openai_client.completions.create(**args_dict)

    def process_output(self, outputs):
        results = []
        for choice in outputs.choices:
            results.append(
                GenerateResult(
                    completion_string=choice.text,
                    completion_tokens=[],  # Not part of the OpenAI spec
                    stop_reason=choice.finish_reason,
                )
            )

        return GenerateResults(results=results)



# SPDX-License-Identifier: Apache-2.0

"""
Test cases for io_adapters/citations.py
"""

# Standard
import datetime

# Third Party
import pytest

# Local
from granite_io import make_io_processor
from granite_io.backend.vllm_server import LocalVLLMServer
from granite_io.io.citations import CitationsCompositeIOProcessor, CitationsIOProcessor
from granite_io.io.granite_3_2.input_processors.granite_3_2_input_processor import (
    Granite3Point2Inputs,
    override_date_for_testing,
)
from granite_io.types import GenerateResult, GenerateResults

_EXAMPLE_CHAT_INPUT = Granite3Point2Inputs.model_validate(
    {
        "messages": [
            {
                "role": "user",
                "content": "What is the visibility level of Git Repos and Issue \
Tracking projects?",
            },
            {
                "role": "assistant",
                "content": "Git Repos and Issue Tracking projects can have one of the \
following visibility levels: private, internal, or public. Private projects are \
visible only to project members, internal projects are visible to all users that are \
logged in to IBM Cloud, and public projects are visible to anyone. By default, \
new projects are set to private visibility level, which is the most secure for your \
data.",
            },
        ],
        "documents": [
            {
                # Original text
                "text": "Git Repos and Issue Tracking is an IBM-hosted component of \
the Continuous Delivery service. All of the data that you provide to Git Repos and \
Issue Tracking, including but not limited to source files, issues, pull requests, and \
project configuration properties, is managed securely within Continuous Delivery. \
However, Git Repos and Issue Tracking supports various mechanisms for exporting, \
sending, or otherwise sharing data to users and third parties. The ability of Git \
Repos and Issue Tracking to share information is typical of many social coding \
platforms. However, such sharing might conflict with regulatory controls that \
apply to your business. After you create a project in Git Repos and Issue Tracking, \
but before you entrust any files, issues, records, or other data with the project, \
review the project settings and change any settings that you deem necessary to \
protect your data. Settings to review include visibility levels, email notifications, \
integrations, web hooks, access tokens, deploy tokens, and deploy keys. Project \
visibility levels \n\nGit Repos and Issue Tracking projects can have one of the \
following visibility levels: private, internal, or public. * Private projects are \
visible only to project members. This setting is the default visibility level for new \
projects, and is the most secure visibility level for your data. * Internal projects \
are visible to all users that are logged in to IBM Cloud. * Public projects are \
visible to anyone. To limit project access to only project members, complete the \
following steps:\n\n\n\n1. From the project sidebar, click Settings > General. \
2. On the General Settings page, click Visibility > project features > permissions. \
3. Locate the Project visibility setting. 4. Select Private, if it is not already \
selected. 5. Click Save changes. Project membership \n\nGit Repos and Issue Tracking \
is a cloud hosted social coding environment that is available to all Continuous \
Delivery users. If you are a Git Repos and Issue Tracking project Maintainer or Owner, \
you can invite any user and group members to the project. IBM Cloud places no \
restrictions on who you can invite to a project."
            },
            {
                "text": "After you create a project in Git Repos and Issue Tracking, \
but before you entrust any files, issues, records, or other data with the project, \
review the project settings and change any settings that are necessary to protect your \
data. \
Settings to review include visibility levels, email notifications, integrations, web \
hooks, access tokens, deploy tokens, and deploy keys. Project visibility levels \
\n\nGit Repos and Issue Tracking projects can have one of the following visibility \
levels: private, internal, or public. * Private projects are visible only to \
project members. This setting is the default visibility level for new projects, and \
is the most secure visibility level for your data. * Internal projects are visible to \
all users that are logged in to IBM Cloud. * Public projects are visible to anyone. \
To limit project access to only project members, complete the following \
steps:\n\n\n\n1. From the project sidebar, click Settings > General. 2. On the \
General Settings page, click Visibility > project features > permissions. 3. Locate \
the Project visibility setting. 4. Select Private, if it is not already selected. \
5. Click Save changes. Project email settings \n\nBy default, Git Repos and Issue \
Tracking notifies project members by way of email about project activities. These \
emails typically include customer-owned data that was provided to Git Repos and Issue \
Tracking by users. For example, if a user posts a comment to an issue, Git Repos and \
Issue Tracking sends an email to all subscribers. The email includes information such \
as a copy of the comment, the user who posted it, and when the comment was posted. \
To turn off all email notifications for your project, complete the following \
steps:\n\n\n\n1. From the project sidebar, click Settings > General. 2. On the \
**General Settings **page, click Visibility > project features > permissions. \
3. Select the Disable email notifications checkbox. 4. Click Save changes. Project \
integrations and webhooks"
            },
        ],
    }
)


def _make_result(content: str):
    """Convenience method to create a fake model output object."""
    return GenerateResult(
        completion_string=content, completion_tokens=[], stop_reason="dummy stop reason"
    )


_TODAYS_DATE = datetime.datetime.now().strftime("%B %d, %Y")


def test_canned_input():
    """
    Validate that the I/O processor handles a single instance of canned input in the
    expected way.
    """
    io_processor = CitationsIOProcessor(None)
    output = io_processor.inputs_to_generate_inputs(_EXAMPLE_CHAT_INPUT)
    print("#####")
    print(output.prompt)
    print("#####")
    expected_prompt = f"""\
<|start_of_role|>system<|end_of_role|>Knowledge Cutoff Date: April 2024.
Today's Date: {_TODAYS_DATE}.
You are Granite, developed by IBM.Write the response to the user's input by strictly aligning with the facts in the provided documents. If the information needed to answer the question is not available in the documents, inform the user that the question cannot be answered based on the available data.<|end_of_text|>
<|start_of_role|>documents<|end_of_role|>Document 0
<c0> Git Repos and Issue Tracking is an IBM-hosted component of the Continuous Delivery service. <c1> All of the data that you provide to Git Repos and Issue Tracking, including but not limited to source files, issues, pull requests, and project configuration properties, is managed securely within Continuous Delivery. <c2> However, Git Repos and Issue Tracking supports various mechanisms for exporting, sending, or otherwise sharing data to users and third parties. <c3> The ability of Git Repos and Issue Tracking to share information is typical of many social coding platforms. <c4> However, such sharing might conflict with regulatory controls that apply to your business. <c5> After you create a project in Git Repos and Issue Tracking, but before you entrust any files, issues, records, or other data with the project, review the project settings and change any settings that you deem necessary to protect your data. <c6> Settings to review include visibility levels, email notifications, integrations, web hooks, access tokens, deploy tokens, and deploy keys. <c7> Project visibility levels 

Git Repos and Issue Tracking projects can have one of the following visibility levels: private, internal, or public. <c8> * Private projects are visible only to project members. <c9> This setting is the default visibility level for new projects, and is the most secure visibility level for your data. <c10> * Internal projects are visible to all users that are logged in to IBM Cloud. <c11> * Public projects are visible to anyone. <c12> To limit project access to only project members, complete the following steps:



1. <c13> From the project sidebar, click Settings > General. <c14> 2. <c15> On the General Settings page, click Visibility > project features > permissions. <c16> 3. <c17> Locate the Project visibility setting. <c18> 4. <c19> Select Private, if it is not already selected. <c20> 5. <c21> Click Save changes. <c22> Project membership 

Git Repos and Issue Tracking is a cloud hosted social coding environment that is available to all Continuous Delivery users. <c23> If you are a Git Repos and Issue Tracking project Maintainer or Owner, you can invite any user and group members to the project. <c24> IBM Cloud places no restrictions on who you can invite to a project.

Document 1
<c25> After you create a project in Git Repos and Issue Tracking, but before you entrust any files, issues, records, or other data with the project, review the project settings and change any settings that are necessary to protect your data. <c26> Settings to review include visibility levels, email notifications, integrations, web hooks, access tokens, deploy tokens, and deploy keys. <c27> Project visibility levels 

Git Repos and Issue Tracking projects can have one of the following visibility levels: private, internal, or public. <c28> * Private projects are visible only to project members. <c29> This setting is the default visibility level for new projects, and is the most secure visibility level for your data. <c30> * Internal projects are visible to all users that are logged in to IBM Cloud. <c31> * Public projects are visible to anyone. <c32> To limit project access to only project members, complete the following steps:



1. <c33> From the project sidebar, click Settings > General. <c34> 2. <c35> On the General Settings page, click Visibility > project features > permissions. <c36> 3. <c37> Locate the Project visibility setting. <c38> 4. <c39> Select Private, if it is not already selected. <c40> 5. <c41> Click Save changes. <c42> Project email settings 

By default, Git Repos and Issue Tracking notifies project members by way of email about project activities. <c43> These emails typically include customer-owned data that was provided to Git Repos and Issue Tracking by users. <c44> For example, if a user posts a comment to an issue, Git Repos and Issue Tracking sends an email to all subscribers. <c45> The email includes information such as a copy of the comment, the user who posted it, and when the comment was posted. <c46> To turn off all email notifications for your project, complete the following steps:



1. <c47> From the project sidebar, click Settings > General. <c48> 2. <c49> On the **General Settings **page, click Visibility > project features > permissions. <c50> 3. <c51> Select the Disable email notifications checkbox. <c52> 4. <c53> Click Save changes. <c54> Project integrations and webhooks<|end_of_text|>
<|start_of_role|>user<|end_of_role|>What is the visibility level of Git Repos and Issue Tracking projects?<|end_of_text|>
<|start_of_role|>assistant<|end_of_role|><r0> Git Repos and Issue Tracking projects can have one of the following visibility levels: private, internal, or public. <r1> Private projects are visible only to project members, internal projects are visible to all users that are logged in to IBM Cloud, and public projects are visible to anyone. <r2> By default, new projects are set to private visibility level, which is the most secure for your data.<|end_of_text|>
<|start_of_role|>system<|end_of_role|>Split the last assistant response into individual sentences. For each sentence in the response, identify the statement IDs from the documents that it references. Ensure that your output includes all response sentence IDs, and for each response sentence ID, provide the corresponding referring document sentence IDs.<|end_of_text|>\
"""  # noqa: E501
    assert output.prompt == expected_prompt


def test_canned_output():
    """
    Validate that the I/O processor handles a single instance of canned model output
    in the expected way.
    """
    io_processor = CitationsIOProcessor(None)

    # Map raw input to just the citation offsets
    raw_output_to_expected = [
        (
            '{"<r0>": ["<c7>"], "<r1>": ["<c8>", "<c10>", "<c11>"]}',
            [
                {
                    "context_begin": 1034,
                    "context_end": 1178,
                    "response_begin": 0,
                    "response_end": 116,
                },
                {
                    "context_begin": 1179,
                    "context_end": 1234,
                    "response_begin": 117,
                    "response_end": 289,
                },
                {
                    "context_begin": 1353,
                    "context_end": 1430,
                    "response_begin": 117,
                    "response_end": 289,
                },
                {
                    "context_begin": 1431,
                    "context_end": 1471,
                    "response_begin": 117,
                    "response_end": 289,
                },
            ],
        ),
        ('{"<r0>": [], "<r1>": []}', []),
        # IO processor currently raises an exception for debugging in this case.
        # ("<invalid model response>", "ERROR"),
    ]

    # Single output
    for raw_output, expected in raw_output_to_expected:
        output = io_processor.output_to_result(
            GenerateResults(results=[_make_result(raw_output)]), _EXAMPLE_CHAT_INPUT
        )
        assert len(output.results) == 1
        # Pull out just the citation offsets for comparison
        citation_offsets = [
            {
                key: getattr(c, key)
                for key in (
                    "context_begin",
                    "context_end",
                    "response_begin",
                    "response_end",
                )
            }
            for c in output.results[0].next_message.citations
        ]
        assert citation_offsets == expected


@pytest.mark.vcr
def test_run_model(lora_server: LocalVLLMServer, fake_date: str):
    """
    Run a chat completion through the LoRA adapter using the I/O processor.
    """
    backend = lora_server.make_lora_backend("citations")
    io_proc = CitationsIOProcessor(backend)
    override_date_for_testing(fake_date)  # For consistent VCR output

    # Pass our example input thorugh the I/O processor and retrieve the result

    chat_result = io_proc.create_chat_completion(_EXAMPLE_CHAT_INPUT)

    # Results vary slightly due to rounding error inside the model.
    assert 4 <= len(chat_result.results[0].next_message.citations) <= 5


@pytest.mark.vcr
def test_run_composite(lora_server: LocalVLLMServer, fake_date: str):
    """
    Run a chat completion through the LoRA adapter using the composite I/O processor.
    """
    granite_backend = lora_server.make_backend()
    lora_backend = lora_server.make_lora_backend("citations")
    granite_io_proc = make_io_processor("Granite 3.2", backend=granite_backend)
    io_proc = CitationsCompositeIOProcessor(granite_io_proc, lora_backend)
    override_date_for_testing(fake_date)  # For consistent VCR output

    # Strip off last message and rerun
    input_without_msg = _EXAMPLE_CHAT_INPUT.model_copy(
        update={"messages": _EXAMPLE_CHAT_INPUT.messages[:-1]}
    ).with_addl_generate_params({"temperature": 0.2, "n": 5})
    results = io_proc.create_chat_completion(input_without_msg)
    assert len(results.results) == 5



# SPDX-License-Identifier: Apache-2.0

"""
This example show how to infer or call a model using the framework
and an Ollama backend to serve the model.
In this scenario, it is a simple text from Wikipedia use case where
citations are enabled.
"""

# Standard
import pprint

# Local
from granite_io import make_backend, make_io_processor
from granite_io.types import ChatCompletionInputs, UserMessage

model_name = "granite3.2:8b"
io_processor = make_io_processor(
    model_name, backend=make_backend("openai", {"model_name": model_name})
)
question = "What team is the most successful hurling team in the last 10 years?"
messages = [UserMessage(content=question)]

# Source: https://en.wikipedia.org/wiki/Limerick_county_hurling_team
documents = [
    {
        "text": "The 2018 season concluded with Limerick winning the 2018 All-Ireland SHC, the team's first since 1973, with a 3–16 to 2–18 point defeat of Galway in the final.The team built on this success, winning the NHL in 2019, 2020 and 2023, the Munster SHC in 2019, 2020, 2021 and 2022 and the All-Ireland SHC again in 2020, 2021 and 2022. Munster Senior Hurling Championship 2023, All Ireland Hurling Championship 2023 to be forever remembered the team to join the Cork hurling Champions of the 40s and the Kilkenny hurling Champions of the 2000s to complete 4 in a row."  # noqa: E501
    },
]

# With RAG and citations
outputs = io_processor.create_chat_completion(
    ChatCompletionInputs(
        messages=messages,
        documents=documents,
        controls={"citations": True},
    )
)
print("\n\n>> Model raw output:\n")
response = outputs.results[0].next_message.raw
print(response)

print("\n\n>> Response:\n")
response = outputs.results[0].next_message.content
print(response)

if outputs.results[0].next_message.citations:
    print("\n\n>> Citations:\n")
    pprint.pprint(outputs.results[0].next_message.citations, sort_dicts=False)

    print("\n\n>> Documents:\n")
    pprint.pprint(outputs.results[0].next_message.documents, sort_dicts=False)

if outputs.results[0].next_message.hallucinations:
    print("\n\n>> Hallucinations:\n")
    pprint.pprint(outputs.results[0].next_message.hallucinations, sort_dicts=False)



# SPDX-License-Identifier: Apache-2.0


"""
I/O processor for the Granite answerability intrinsic.
"""

# Local
from granite_io.io.base import (
    InputOutputProcessor,
    ModelDirectInputOutputProcessorWithGenerate,
)
from granite_io.io.granite_3_2.input_processors.granite_3_2_input_processor import (
    Granite3Point2InputProcessor,
    Granite3Point2Inputs,
)
from granite_io.types import (
    AssistantMessage,
    ChatCompletionInputs,
    ChatCompletionResult,
    ChatCompletionResults,
    GenerateInputs,
    GenerateResults,
)


class AnswerabilityIOProcessor(ModelDirectInputOutputProcessorWithGenerate):
    """
    I/O processor for the answerability intrinsic, also known as the LoRA Adapter for
    Answerability Classification
    Takes as input a chat completion and returns a completion with a 1/0 answerability
    flag as a string in the content field.

    Example raw input:
    ```
    <|start_of_role|>documents<|end_of_role|> ... <|end_of_text|>
    <|start_of_role|>user<|end_of_role|>Can I go tubing along the Russian River in
    Sonoma County?<|end_of_text|> <|start_of_role|>assistant<|end_of_role|>Yes, tubing
    along the Russian River in Sonoma County is a popular activity. The recommended
    route for tubing is the 2-mile stretch from Steelhead Beach to Sunset Beach in
    Forestville.<|end_of_text|> <|start_of_role|>user<|end_of_role|>How can I arrange
    transportation between Steelhead Beach and Sunset Beach?<|end_of_text|>
    <|start_of_role|>answerability<|end_of_role|>
    ```

    Raw output should be either `answerable` or `unanswerable`.
    """

    def __init__(self, backend):
        super().__init__(backend=backend)

        # Input processor for the base model, which does most of the input formatting.
        self.base_input_processor = Granite3Point2InputProcessor()

    def inputs_to_generate_inputs(
        self, inputs: ChatCompletionInputs, add_generation_prompt: bool = True
    ) -> GenerateInputs:
        # Validate the input and convert to Granite input
        inputs = Granite3Point2Inputs.model_validate(inputs.model_dump())

        # Check for the invariants that the model expects its input to satisfy
        if not inputs.documents:
            raise ValueError("Input does not contain documents")
        if not inputs.messages[-1].role == "user":
            raise ValueError("Last message is not a user message")

        # The beginning of the prompt doesn't change relative to base Granite 3.2
        prompt = self.base_input_processor.transform(inputs, False)

        # Only the generation prompt portion changes
        if add_generation_prompt:
            prompt = prompt + "<|start_of_role|>answerability<|end_of_role|>"

        generate_inputs_before = (
            inputs.generate_inputs if inputs.generate_inputs else GenerateInputs()
        )
        result = generate_inputs_before.model_copy(
            update={
                "prompt": prompt,
                # Ensure enough tokens to produce "unanswerable"
                "max_tokens": 16,
                # Enable constrained decoding on vLLM backends
                "extra_body": {"guided_choice": ["answerable", "unanswerable"]},
            }
        )
        return result

    def output_to_result(
        self, output: GenerateResults, inputs: ChatCompletionInputs | None = None
    ) -> ChatCompletionResults:
        results = []
        for raw_result in output.results:
            raw_str = raw_result.completion_string
            if raw_str in ("answerable", "unanswerable"):
                results.append(
                    ChatCompletionResult(
                        next_message=AssistantMessage(
                            content=raw_result.completion_string
                        )
                    )
                )
            # Model may forget to generate end of string
            elif raw_str.startswith("answerable"):
                results.append(
                    ChatCompletionResult(
                        next_message=AssistantMessage(content="answerable", raw=raw_str)
                    )
                )
            elif raw_str.startswith("unanswerable"):
                results.append(
                    ChatCompletionResult(
                        next_message=AssistantMessage(
                            content="unanswerable", raw=raw_str
                        )
                    )
                )
            else:
                # Improper output. Pass through as an error message for now instead of
                # raising an exception from the I/O processor.
                results.append(
                    ChatCompletionResult(
                        next_message=AssistantMessage(content="ERROR", raw=raw_str),
                    )
                )
        return ChatCompletionResults(results=results)


DEFAULT_CANNED_RESPONSE = (
    "Sorry, but I am unable to answer this question from the documents retrieved."
)


class AnswerabilityCompositeIOProcessor(InputOutputProcessor):
    """
    Composite I/O processor that checks a RAG request for answerability with the
    target document snippets and triggers an appropriate fallback if the request
    is unanswerable with the target documents.
    """

    def __init__(
        self,
        generator: InputOutputProcessor,
        answerability: InputOutputProcessor,
        fallback_type: str = "canned_response",
        canned_response: str = DEFAULT_CANNED_RESPONSE,
    ):
        """
        :param generator: I/O processor that generates the results that this I/O
         processor shoid validate.
        :param answerability_io_proc: IO processor for the answerability model.
         Should return either "answerable" or "unanswerable".
        :param fallback_type: One of "canned_response" or "generate_without_docs"
        :param canned_response: Fallback response to use if ``fallback_type`` is
         "canned_response"
        """
        if fallback_type not in ("canned_response", "generate_without_docs"):
            raise ValueError(
                f"Unknown fallback type '{fallback_type}'. Should be one "
                f"of 'canned_response' or "
                "enerate_without_docs"
            )
        self._generator = generator
        self._answerability = answerability
        self._fallback_type = fallback_type
        self._canned_response = canned_response

    def update_threshold(self, threshold: float):
        """Convenience method to update the filtering threshold.

        :param threshold: New value to be applied to subsequent calls to the certainty
         I/O processor."""
        self._threshold = threshold

    async def acreate_chat_completion(
        self, inputs: ChatCompletionInputs
    ) -> ChatCompletionResults:
        # Run a single answerability check
        answerability_output = (
            await self._answerability.acreate_chat_completion(
                inputs.with_addl_generate_params({"temperature": 0.0, "n": 1})
            )
        ).results[0]

        if answerability_output.next_message.content == "answerable":
            # User message is answerable. Return a result.
            return await self._generator.acreate_chat_completion(inputs)
        if self._fallback_type == "generate_without_docs":
            # User message is not answerable; attempt to generate a non-RAG result.
            return await self._generator.acreate_chat_completion(
                inputs.model_copy(update={"documents": None})
            )
        # if unanswerable and self._fallback_type == "canned_response":
        return ChatCompletionResults(
            results=[
                ChatCompletionResult(
                    next_message=AssistantMessage(content=self._canned_response)
                )
            ]
        )



# SPDX-License-Identifier: Apache-2.0

# Standard
from pathlib import Path
import datetime
import re


def load_text_file(file_name: str) -> str:
    text = Path(file_name).read_text(encoding="UTF-8")
    return text


def fix_granite_date(prompt_with_wrong_date: str) -> str:
    """
    Replace the date string in a Granite prompt with today's date, in Granite prompts'
    date format
    """
    date_str = datetime.datetime.now().strftime("%B %d, %Y")
    return re.sub(
        r"Today's Date: \w+ \d\d, \d\d\d\d\.",
        f"Today's Date: {date_str}.",
        prompt_with_wrong_date,
    )



# SPDX-License-Identifier: Apache-2.0


# Standard
from typing import TYPE_CHECKING

# Local
from granite_io.backend.base import Backend
from granite_io.backend.registry import backend
from granite_io.optional import import_optional
from granite_io.types import GenerateInputs

if TYPE_CHECKING:
    # Third Party
    import litellm  # noqa: F401


@backend(
    "litellm",
    config_schema={
        "properties": {
            "model_name": {"type": "string"},
            # "openai_base_url": {"type": "string"},
            # "openai_api_key": {"type": "string"},
        }
    },
)
class LiteLLMBackend(Backend):
    async def generate(self, inputs: GenerateInputs):
        """Run a direct /completions call"""

        with import_optional("litellm"):
            # Third Party
            import litellm

        return await litellm.atext_completion(**inputs.model_dump())



# SPDX-License-Identifier: Apache-2.0

"""
Common shared types
"""

# Standard
from collections.abc import Mapping
from typing import Any, List, Optional, Union

# Third Party
from typing_extensions import Literal, TypeAlias
import pydantic


class FunctionCall(pydantic.BaseModel):
    id: str | None = None
    name: str

    # This field should adhere to the argument schema from the  associated
    # FunctionDefinition in the generation request that produced it.
    arguments: dict[str, Any] | None


class Hallucination(pydantic.BaseModel):
    """Hallucination data as returned by the model output parser"""

    hallucination_id: str
    risk: str
    response_text: str
    response_begin: int
    response_end: int


class Citation(pydantic.BaseModel):
    """Citation data as returned by the model output parser"""

    citation_id: str
    doc_id: str
    context_text: str
    context_begin: int
    context_end: int
    response_text: str
    response_begin: int
    response_end: int


class Document(pydantic.BaseModel):
    """Document data as returned by the model output parser"""

    doc_id: str
    text: str


class _ChatMessageBase(pydantic.BaseModel):
    """Base class for all message types.

    Due to the vaguaries of Pydantic's JSON parser, we use this class only for common
    functionality, and NOT for defining a common dataclass base type. Use the
    :class:`ChatMessage` type alias to annotate a field or argument as accepting all
    subclasses of this one."""

    content: str
    """Every message has raw string content, even if it also contains parsed structured
    content such as a JSON record."""

    def to_openai_json(self):
        result = {"role": self.role, "content": self.content}
        return result


class UserMessage(_ChatMessageBase):
    role: Literal["user"] = "user"


class AssistantMessage(_ChatMessageBase):
    role: Literal["assistant"] = "assistant"
    tool_calls: list[FunctionCall] = []
    reasoning_content: str | None = None
    # Raw response content without any parsing for re-serialization
    _raw: str | None = None
    citations: list[Citation] | None = None
    documents: list[Document] | None = None
    hallucinations: list[Hallucination] | None = None
    stop_reason: str | None = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._raw = kwargs.pop("raw", None)

    @property
    def raw(self) -> str:
        """Get the raw content of the response"""
        return self._raw if self._raw is not None else self.content


class ToolResultMessage(_ChatMessageBase):
    role: Literal["tool"] = "tool"
    tool_call_id: str


class SystemMessage(_ChatMessageBase):
    role: Literal["system"] = "system"


ChatMessage: TypeAlias = (
    UserMessage | AssistantMessage | ToolResultMessage | SystemMessage
)
"""Type alias for all message types. We use this Union instead of the actual base class
:class:`_ChatMessageBase` so that Pydantic can parse the message list from JSON."""


class FunctionDefinition(pydantic.BaseModel):
    name: str
    description: str | None = None

    # This field holds a JSON schema for a record, but the `jsonschema` package doesn't
    # define an object type for such a schema, instead using a dictionary.
    parameters: dict[str, Any] | None = None

    def to_openai_json(self) -> dict:
        """
        :returns: JSON representation of this function, as the Python equivalent of the
            standard JSON used in the OpenAI chat completions API for the `tools`
            argument.
        """
        raise NotImplementedError("TODO: Implement this")


class GenerateInputs(pydantic.BaseModel):
    """Common inputs for backends

    Attributes:

        OPTIONAL PARAMS
            prompt: The prompt(s) to generate completions for.
            model: Model name or ID.
            best_of: Generates best_of completions server-side.
            echo: Echo back the prompt in addition to the completion.
            frequency_penalty: Penalize new tokens based on their existing frequency.
            logit_bias: Modify the likelihood of specified tokens.
            logprobs: Include the log probabilities on the most likely tokens.
            max_tokens: The maximum number of tokens to generate in the completion.
            n: How many completions to generate for each prompt.
            presence_penalty: Penalize new tokens based on whether they are in the text.
            stop: Sequences where the API will stop generating further tokens.
            stream: Whether to stream back partial progress.
            stream_options: A dictionary containing options for the streaming response.
            suffix: The suffix that comes after a completion of inserted text.
            temperature: The temperature parameter for controlling randomness of output.
            top_p: The top-p parameter for nucleus sampling.
            user: A unique identifier representing your end-user.
            extra_headers: Additional headers to include in the request.
    """

    prompt: Optional[Union[str, List[Union[str, List[Union[str, List[int]]]]]]] = None
    model: Optional[str] = None
    best_of: Optional[int] = None
    echo: Optional[bool] = None
    frequency_penalty: Optional[float] = None
    logit_bias: Optional[dict] = None
    logprobs: Optional[Union[int, bool]] = None
    max_tokens: Optional[int] = None
    n: Optional[int] = None
    presence_penalty: Optional[float] = None
    stop: Union[Optional[str], List[str], None] = None
    stream: Optional[bool] = None
    stream_options: Optional[dict] = None
    suffix: Optional[str] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    user: Optional[str] = None
    extra_headers: Optional[Mapping[str, str]] = None
    extra_body: Optional[Mapping[str, Any]] = None

    model_config = pydantic.ConfigDict(
        # Pass through arbitrary additional keyword arguments for handling by model- or
        # specific I/O processors.
        arbitrary_types_allowed=True,
        extra="allow",
    )


class ChatCompletionInputs(pydantic.BaseModel):
    """
    Class that represents the lowest-common-denominator inputs to a chat completion
    call.  Individual input/output processors can extend this schema with their own
    additional proprietary fields.
    """

    messages: list[ChatMessage]
    tools: list[FunctionDefinition] = []
    generate_inputs: Optional[GenerateInputs] = None
    model_config = pydantic.ConfigDict(
        # Pass through arbitrary additional keyword arguments for handling by model- or
        # specific I/O processors.
        extra="allow"
    )

    def __getattr__(self, name: str) -> any:
        """Allow attribute access for unknown attributes"""
        try:
            return super().__getattr__(name)
        except AttributeError:
            return None

    def with_next_message(self, next_message: ChatMessage) -> "ChatCompletionInputs":
        """
        :param next_message: Additional message to add to the conversation

        :returns: a version of this object with one additional message in the messages
        list. Does not modify the original object.
        """
        new_messages = self.messages.copy()
        new_messages.append(next_message)
        return self.model_copy(update={"messages": new_messages})

    def with_addl_generate_params(
        self, params: Mapping[str, object]
    ) -> "ChatCompletionInputs":
        """
        :param params: Additional parameters to add.

        :returns: a version of this object with one additional message in the messages
        list. Does not modify the original object.
        Leaves in place any generation parameters already present.
        """
        previous_inputs = (
            self.generate_inputs
            if self.generate_inputs is not None
            else GenerateInputs()
        )
        new_inputs = previous_inputs.model_copy(update=params)
        return self.model_copy(update={"generate_inputs": new_inputs})


class ChatCompletionResult(pydantic.BaseModel):
    """
    Class that represents the lowest-common-denominator outputs of a chat completion
    call.  Individual input/output processors can extend this schema with their own
    additional proprietary fields.
    """

    next_message: ChatMessage

    def __getattr__(self, name: str) -> any:
        """Allow attribute access for unknown attributes"""
        try:
            return super().__getattr__(name)
        except AttributeError:
            return None


class ChatCompletionResults(pydantic.BaseModel):
    results: list[ChatCompletionResult]


class GenerateResult(pydantic.BaseModel):
    """
    All the things that our internal :func:`generate()` methods return,
    rolled into a dataclass for ease of maintenance.
    """

    # Not including input characters
    completion_string: str

    # Not including input tokens
    completion_tokens: list[int]

    stop_reason: str


class GenerateResults(pydantic.BaseModel):
    results: list[GenerateResult]



# SPDX-License-Identifier: Apache-2.0

"""
Widget for visualizing citations.
"""

# Standard
import pathlib

# Third Party
import anywidget
import traitlets

# Local
from granite_io.types import (
    ChatCompletionInputs,
    ChatCompletionResults,
)


class CitationsWidget:
    def show(self, inputs: ChatCompletionInputs, outputs: ChatCompletionResults):
        documents = [doc.model_dump(mode="json") for doc in inputs.documents]
        response = outputs.results[0].next_message.content
        citations = [
            citation.model_dump(mode="json")
            for citation in outputs.results[0].next_message.citations
        ]

        return CitationsWidgetInstance(
            response=response, documents=documents, citations=citations
        )


class CitationsWidgetInstance(anywidget.AnyWidget):
    _esm = pathlib.Path(__file__).parent / "index.js"
    _css = pathlib.Path(__file__).parent / "index.css"

    response = traitlets.Unicode({}).tag(sync=True)
    documents = traitlets.List([]).tag(sync=True)
    citations = traitlets.List([]).tag(sync=True)



# SPDX-License-Identifier: Apache-2.0

# Standard
import json
import uuid

# Third Party
import pydantic

# Local
from granite_io.io.base import OutputProcessor
from granite_io.io.consts import (
    _GRANITE_3_3_2B_HF,
    _GRANITE_3_3_2B_OLLAMA,
    _GRANITE_3_3_8B_HF,
    _GRANITE_3_3_8B_OLLAMA,
    _GRANITE_3_3_COT_END,
    _GRANITE_3_3_COT_START,
    _GRANITE_3_3_MODEL_NAME,
    _GRANITE_3_3_RESP_END,
    _GRANITE_3_3_RESP_START,
)
from granite_io.io.granite_3_3.input_processors.granite_3_3_input_processor import (
    Granite3Point3Inputs,
)
from granite_io.io.granite_3_3.output_processors.granite_3_3_output_parser import (
    parse_model_output,
)
from granite_io.io.registry import output_processor
from granite_io.types import (
    AssistantMessage,
    ChatCompletionInputs,
    ChatCompletionResult,
    ChatCompletionResults,
    FunctionCall,
    GenerateResults,
)

# Some versions of the model are known to shorten "Here is" to "Here's", so we
# provide alternate forms of these strings for those versions.
_COT_START_ALTERNATIVES = [
    _GRANITE_3_3_COT_START,
]
_COT_END_ALTERNATIVES = [
    _GRANITE_3_3_COT_END,
]

_MODEL_NAME = "Granite 3.3"


def _random_uuid() -> str:
    """:returns: hexadecimal data suitable to use as a unique identifier"""
    return str(uuid.uuid4())


@output_processor(
    _GRANITE_3_3_MODEL_NAME,
    # Huggingface
    _GRANITE_3_3_2B_HF,
    _GRANITE_3_3_8B_HF,
    # Ollama
    "granite3.3",
    _GRANITE_3_3_2B_OLLAMA,
    _GRANITE_3_3_8B_OLLAMA,
)
class Granite3Point3OutputProcessor(OutputProcessor):
    """
    Output processor for version 3.3 of the main Granite models, all sizes.
    """

    def transform(
        self,
        output: GenerateResults,
        inputs: ChatCompletionInputs | None = None,
    ) -> ChatCompletionResults:
        # Downcast to a Granite-specific request type with possible additional fields.
        # This operation also performs additional validation.
        inputs = Granite3Point3Inputs.model_validate(inputs.model_dump())

        results = []
        for result in output.results:
            output = result.completion_string
            original_output = output

            # Parse out CoT reasoning
            cot = None
            if inputs.thinking:
                cot_start_span = None
                cot_end_span = None
                for cot_start_str in _COT_START_ALTERNATIVES:
                    if (cot_start_pos := output.find(cot_start_str)) != -1:
                        cot_start_span = (
                            cot_start_pos,
                            cot_start_pos + len(cot_start_str),
                        )
                        break
                for cot_end_str in _COT_END_ALTERNATIVES:
                    if (cot_end_pos := output.find(cot_end_str)) != -1:
                        cot_end_span = (cot_end_pos, cot_end_pos + len(cot_end_str))
                        break

                if (
                    cot_start_span
                    and cot_end_span
                    and cot_end_span[0] > cot_start_span[1]
                ):
                    cot = output[cot_start_span[1] : cot_end_span[0]].strip()
                    output = (
                        output[: cot_start_span[0]] + output[cot_end_span[1] :].strip()
                    )
                    resp_start_span = None
                    resp_end_span = None
                    for resp_start_str in [_GRANITE_3_3_RESP_START]:
                        if (resp_start_pos := output.find(resp_start_str)) != -1:
                            resp_start_span = (
                                resp_start_pos,
                                resp_start_pos + len(resp_start_str),
                            )
                            break
                    for resp_end_str in [_GRANITE_3_3_RESP_END]:
                        if (resp_end_pos := output.find(resp_end_str)) != -1:
                            resp_end_span = (
                                resp_end_pos,
                                resp_end_pos + len(resp_end_str),
                            )
                            break
                    if (
                        resp_start_span
                        and resp_end_span
                        and resp_end_span[0] > resp_start_span[1]
                    ):
                        output = output[resp_start_span[1] : resp_end_span[0]].strip()

            # Parse out tool calls
            tool_calls = []
            if inputs.tools and output.startswith("<tool_call>"):
                # Basic tool call parsing: assume well-formed JSON that adheres to the
                # argument schema specified in the request.
                try:
                    tool_calls = json.loads(output[len("<tool_call>") :])
                    if not isinstance(tool_calls, list):
                        raise TypeError("Model didn't output a list of tool calls")
                    tool_calls = [
                        FunctionCall.model_validate(tool_call_json)
                        for tool_call_json in tool_calls
                    ]
                    for function_call in tool_calls:
                        # Model may decide not to produce IDs
                        if function_call.id is None:
                            function_call.id = _random_uuid()
                    # Output has been turned to tools
                    output = ""
                except (ValueError, TypeError, pydantic.ValidationError):
                    # Parsing failed; flow through
                    pass

            # Parse out citations, documents and hallucinations
            try:
                parsed_output = parse_model_output(output, inputs.documents)
            except Exception as err:
                raise ValueError(
                    "Failed to parse citations, documents and hallucinations "
                    "from model output."
                ) from err

            results.append(
                ChatCompletionResult(
                    next_message=AssistantMessage(
                        citations=parsed_output["citations"],
                        content=(parsed_output["response"]),
                        documents=parsed_output["docs"],
                        hallucinations=parsed_output["hallucinations"],
                        reasoning_content=cot,
                        raw=original_output,
                        stop_reason=result.stop_reason,
                        tool_calls=tool_calls,
                    )
                )
            )

        return ChatCompletionResults(results=results)



# SPDX-License-Identifier: Apache-2.0

# Note: Never rename this file to "vllm.py". You will have a bad time.

# Standard
from collections.abc import Iterable
import asyncio
import logging
import os
import shutil
import signal
import socketserver
import subprocess
import sys
import time
import urllib
import uuid

# Third Party
import aconfig
import aiohttp

# Local
from granite_io.backend.openai import OpenAIBackend

# Perform the "set sensible defaults for Python logging" ritual.
logger = logging.getLogger("granite_io.backend.vllm_server")
logger.setLevel(logging.INFO)
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(
    logging.Formatter("%(levelname)s %(asctime)s %(message)s", datefmt="%H:%M:%S")
)
logger.addHandler(handler)


class LocalVLLMServer:
    """
    Class that manages a vLLM server subprocess on the local machine.
    """

    def __init__(
        self,
        model_name: str,
        device_name: str = "auto",
        served_model_name: str | None = None,
        api_key: str | None = None,
        port: int | None = None,
        gpu_memory_utilization: float = 0.45,
        max_model_len: int = 32768,
        enforce_eager: bool = True,
        log_level: str = "INFO",
        lora_adapters: Iterable[tuple[str, str]] = tuple(),
        max_lora_rank: int = 64,
    ):
        """
        :param model_name: Path to local file or Hugging Face coordinates of model
        :param served_model_name: Optional alias under which the model should be named
         in the OpenAI API
        :param device_name: Name of vLLM device to use for inference. Current options
         are {auto,cuda,neuron,cpu,tpu,xpu,hpu}.
         Note that the "cpu" option requires building vLLM from source.
        :param api_key: Optional API key for the server to require. Otherwise this class
         will generate a random key.
        :param port: Optional port on localhost to use. If not specified, this class
         will pick a random unused port.
        :param gpu_memory_utilization: What fraction of the GPU's memory to dedicate
         to the target model
        :param max_model_len: Maximum context length to use before truncating to avoid
         running out of GPU memory.
        :param enforce_eager: If ``True`` skip compilation to make the server start up
         faster.
        :param log_level: Logging level for the vLLM subprocess
        :param lora_adapters: Map from model name to LoRA adapter location
        :param max_lora_rank: vLLM needs you to specify an upper bound on the size of
         the shared dimension of the low-rank approximations in LoRA adapters.
        """
        self._model_name = model_name
        self._served_model_name = served_model_name
        self._lora_adapters = lora_adapters
        self._lora_names = [t[0] for t in lora_adapters]

        vllm_exec = shutil.which("vllm")
        if vllm_exec is None:
            raise ValueError("vLLM not installed.")

        if not port:
            # Find an open port on localhost
            with socketserver.TCPServer(("localhost", 0), None) as s:
                port = s.server_address[1]
        self._server_port = port

        # Generate shared secret so that other local processes can't hijack our server.
        self._api_key = api_key if api_key else str(uuid.uuid4())

        environment = os.environ.copy()
        # Disable annoying log messages about current throughput being zero
        # Unfortunately the only documented way to do this is to turn off all
        # logging.
        # TODO: Look for undocumented solutions.
        environment["VLLM_LOGGING_LEVEL"] = log_level
        environment["VLLM_API_KEY"] = self._api_key

        # Immediately start up a server process on the open port
        command_parts = [
            vllm_exec,
            "serve",
            model_name,
            "--port",
            str(self._server_port),
            "--gpu-memory-utilization",
            str(gpu_memory_utilization),
            "--max-model-len",
            str(max_model_len),
            "--guided_decoding_backend",
            "outlines",
            "--device",
            device_name,
        ]
        if enforce_eager:
            command_parts.append("--enforce-eager")
        if served_model_name is not None:
            command_parts.append("--served-model-name")
            command_parts.append(served_model_name)
        if len(lora_adapters) > 0:
            command_parts.append("--enable-lora")
            command_parts.append("--max_lora_rank")
            command_parts.append(str(max_lora_rank))
            command_parts.append("--lora-modules")
            for k, v in lora_adapters:
                command_parts.append(f"{k}={v}")

        logger.info("Running: %s", " ".join(command_parts))  # pylint: disable=logging-not-lazy
        self._subproc = subprocess.Popen(command_parts, env=environment)  # pylint: disable=consider-using-with

    def __repr__(self):
        return f"LocalVLLMServer({self._model_name} -> {self._base_url()})"

    def _base_url(self):
        return f"http://localhost:{self._server_port}"

    @property
    def openai_url(self) -> str:
        return f"{self._base_url()}/v1"

    @property
    def openai_api_key(self) -> str:
        return self._api_key

    def wait_for_startup(self, timeout_sec: float | None = None):
        """
        Blocks  until the server has started.
        :param timeout_sec: Optional upper limit for how long to block. If this
         limit is reached, this method will raise a TimeoutError
        """
        start_sec = time.time()
        while timeout_sec is None or time.time() - start_sec < timeout_sec:
            try:  # Exceptions as control flow due to library design
                with urllib.request.urlopen(self._base_url() + "/ping") as response:
                    _ = response.read().decode("utf-8")
                return  # Success
            except urllib.error.URLError:
                time.sleep(1)
        raise TimeoutError(
            f"Failed to connect to {self._base_url()} after {timeout_sec} seconds."
        )

    async def await_for_startup(self, timeout_sec: float | None = None):
        """
        Blocks the local coroutine until the server has started.
        :param timeout_sec: Optional upper limit for how long to block. If this
         limit is reached, this method will raise a TimeoutError
        """
        start_sec = time.time()
        while timeout_sec is None or time.time() - start_sec < timeout_sec:
            try:  # Exceptions as control flow due to aiohttp library design
                async with (
                    aiohttp.ClientSession() as session,
                    session.get(self._base_url() + "/ping") as resp,
                ):
                    await resp.text()
                return  # Success
            except (ConnectionRefusedError, aiohttp.ClientConnectorError):
                await asyncio.sleep(1)
        raise TimeoutError(
            f"Failed to connect to {self._base_url()} after {timeout_sec} seconds."
        )

    def shutdown(self):
        # Sending SIGINT to the vLLM process seems to be the only way to stop it.
        # DO NOT USE SIGKILL!!!
        self._subproc.send_signal(signal.SIGINT)

    def make_backend(self) -> OpenAIBackend:
        """
        :returns: A backend instance pointed at the primary model that our subprocess
         is serving.
        """
        return OpenAIBackend(
            aconfig.Config(
                {
                    "model_name": (
                        self._served_model_name
                        if self._served_model_name
                        else self._model_name
                    ),
                    "openai_base_url": f"{self._base_url()}/v1",
                    "openai_api_key": self._api_key,
                }
            )
        )

    def make_lora_backend(self, lora_name: str) -> OpenAIBackend:
        """
        :param lora_name: Name of one of the LoRA adapters that was passed to the
        constructor of this object.

        :returns: A backend instance pointed at the specified LoRA adapter.
        """
        if lora_name not in self._lora_names:
            raise ValueError(
                f"Unexpected LoRA adapter name {lora_name}. Known names "
                f"are: {self._lora_names}"
            )
        return OpenAIBackend(
            aconfig.Config(
                {
                    "model_name": lora_name,
                    "openai_base_url": f"{self._base_url()}/v1",
                    "openai_api_key": self._api_key,
                }
            )
        )



# SPDX-License-Identifier: Apache-2.0

"""
Tests for the majority voting I/O processor
"""

# Third Party
from litellm import UnsupportedParamsError
import pytest

# Local
from granite_io import make_io_processor
from granite_io.backend import Backend
from granite_io.backend.litellm import LiteLLMBackend
from granite_io.backend.transformers import TransformersBackend
from granite_io.io.consts import _GRANITE_3_2_MODEL_NAME
from granite_io.io.voting import (
    MajorityVotingProcessor,
    MBRDMajorityVotingProcessor,
    integer_normalizer,
)
from granite_io.types import ChatCompletionInputs, ChatCompletionResults


@pytest.mark.vcr(record_mode="new_episodes")
@pytest.mark.block_network
def test_numeric_voting(backend_x: Backend):
    if isinstance(backend_x, TransformersBackend):
        pytest.xfail(
            "TransformersBackend top-k currently returning low-quality results"
        )

    # At the moment, git LFS is broken on the current repo, so we have reduced the
    # number of samples to keep the cassette file size small.
    SAMPLES_PER_COMPLETION = 3

    base_processor = make_io_processor(_GRANITE_3_2_MODEL_NAME, backend=backend_x)
    voting_processor = MajorityVotingProcessor(
        base_processor,
        integer_normalizer,
        samples_per_completion=SAMPLES_PER_COMPLETION,
    )

    first_number = 1
    second_number = 22
    completion_inputs = ChatCompletionInputs(
        messages=[
            {
                "role": "user",
                "content": f"What is {first_number} + {second_number}?\n"
                f"Answer with just a number please.",
            }
        ],
        thinking=True,
        num_return_sequences=1,
    )
    try:
        results = voting_processor.create_chat_completion(completion_inputs)
    except UnsupportedParamsError:
        # Known issue with LiteLLMBackend
        if isinstance(backend_x, LiteLLMBackend):
            pytest.xfail("LiteLLMBackend support for n > 1 varies by provider")

    assert isinstance(results, ChatCompletionResults)
    assert len(results.results) == 1

    # Due to the git LFS workaround described above, this test case doesn't reliably
    # generate the correct result here. If we enable git LFS support, we should
    # re-enable the following assertion.
    # assert (int(results.results[0].next_message.content)
    #         == first_number + second_number)


@pytest.mark.vcr(record_mode="new_episodes")
@pytest.mark.block_network
def test_mbrd_majority_voting(backend_x: Backend):
    base_processor = make_io_processor(_GRANITE_3_2_MODEL_NAME, backend=backend_x)
    voting_processor = MBRDMajorityVotingProcessor(
        base_processor,
    )

    first_number = 1
    second_number = 22

    if isinstance(backend_x, TransformersBackend):
        generate_inputs = {"n": 10, "max_tokens": 1024}
    else:
        generate_inputs = {"n": 10, "temperature": 0.6, "max_tokens": 1024}
    completion_inputs = ChatCompletionInputs(
        messages=[
            {
                "role": "user",
                "content": f"What is {first_number} + {second_number}?\n"
                f"Answer with just a number please.",
            }
        ],
        thinking=True,
        generate_inputs=generate_inputs,
    )
    try:
        results = voting_processor.create_chat_completion(completion_inputs)
    except UnsupportedParamsError:
        # Known issue with LiteLLMBackend
        if isinstance(backend_x, LiteLLMBackend):
            pytest.xfail("LiteLLMBackend support for n > 1 varies by provider")

    assert isinstance(results, ChatCompletionResults)
    assert len(results.results) == 1
    assert results.results[0].next_message.content in (
        "The sum of 1 and 22 is 23.",
        "23",
    )



# SPDX-License-Identifier: Apache-2.0

"""
Tests for the core types
"""

# Third Party
import pytest

# Local
from granite_io import types


def test_function_call():
    fc = types.FunctionCall(id="1", name="test_func", arguments={"arg1": "value"})
    assert fc.id == "1"
    assert fc.name == "test_func"
    assert fc.arguments == {"arg1": "value"}


def test_chat_message_types():
    um = types.UserMessage(content="user content")
    am = types.AssistantMessage(content="assistant content", tool_calls=[])
    trm = types.ToolResultMessage(content="tool result content", tool_call_id="123")
    sm = types.SystemMessage(content="system content")

    assert um.role == "user"
    assert am.role == "assistant"
    assert trm.role == "tool"
    assert sm.role == "system"
    assert um.to_openai_json()["role"] == "user"
    assert am.to_openai_json()["role"] == "assistant"
    assert trm.to_openai_json()["role"] == "tool"
    assert sm.to_openai_json()["role"] == "system"


def test_function_definition():
    fd = types.FunctionDefinition(name="test_func", description="Test function")
    assert fd.name == "test_func"
    assert fd.description == "Test function"
    assert fd.parameters is None

    with pytest.raises(NotImplementedError):
        fd.to_openai_json()


def test_chat_completion_inputs():
    cci = types.ChatCompletionInputs(messages=[])
    assert len(cci.messages) == 0
    assert len(cci.tools) == 0

    # Setting additional attributes should work as expected
    cci.additional_attr = "value"
    assert hasattr(cci, "additional_attr") is True
    assert cci.additional_attr == "value"

    # Getting unknown attributes should return None
    assert cci.foobar is None


def test_chat_completion_result():
    ccrr = types.ChatCompletionResult(
        next_message=types.UserMessage(content="test message")
    )
    assert hasattr(ccrr, "next_message") is True

    # Testing __getattr__ behavior for unknown attributes
    assert ccrr.foobar is None
    assert ccrr.bazbat is None


def test_generate_result():
    gr = types.GenerateResult(
        completion_string="generated string",
        completion_tokens=[1, 2, 3],
        stop_reason="stop reason",
    )
    assert gr.completion_string == "generated string"
    assert gr.completion_tokens == [1, 2, 3]
    assert gr.stop_reason == "stop reason"



# SPDX-License-Identifier: Apache-2.0

"""
Common shared constants
"""

# Granite 3.2 constants
_GRANITE_3_2_COT_START = "Here is my thought process:"
_GRANITE_3_2_COT_END = "Here is my response:"

_GRANITE_3_2_2B_OLLAMA = "granite3.2:2b"
_GRANITE_3_2_2B_HF = "ibm-granite/granite-3.2-2b-instruct"
_GRANITE_3_2_MODEL_NAME = "Granite 3.2"

# Granite 3.3 constants
_GRANITE_3_3_MODEL_NAME = "Granite 3.3"
_GRANITE_3_3_2B_OLLAMA = "granite3.3:2b"
_GRANITE_3_3_8B_OLLAMA = "granite3.3:8b"
_GRANITE_3_3_2B_HF = "ibm-granite/granite-3.3-2b-instruct"
_GRANITE_3_3_8B_HF = "ibm-granite/granite-3.3-8b-instruct"

_GRANITE_3_3_CITATIONS_START = '{"id": "citation"}'
_GRANITE_3_3_HALLUCINATIONS_START = '{"id": "hallucination"}'
_GRANITE_3_3_CITE_START = "<|start_of_cite|>"
_GRANITE_3_3_CITE_END = "<|end_of_cite|>"
_GRANITE_3_3_COT_START = "<think>"
_GRANITE_3_3_COT_END = "</think>"
_GRANITE_3_3_RESP_START = "<response>"
_GRANITE_3_3_RESP_END = "</response>"



# SPDX-License-Identifier: Apache-2.0

"""
I/O processor for the Granite certainty intrinsic.

See model card at https://huggingface.co/ibm-granite/granite-uncertainty-3.2-8b-lora
"""

# Local
from granite_io.backend.base import Backend
from granite_io.io.base import (
    InputOutputProcessor,
    ModelDirectInputOutputProcessorWithGenerate,
)
from granite_io.io.granite_3_2.input_processors.granite_3_2_input_processor import (
    Granite3Point2InputProcessor,
    Granite3Point2Inputs,
)
from granite_io.types import (
    AssistantMessage,
    ChatCompletionInputs,
    ChatCompletionResult,
    ChatCompletionResults,
    GenerateInputs,
    GenerateResults,
)


class CertaintyIOProcessor(ModelDirectInputOutputProcessorWithGenerate):
    """
    I/O processor for the certainty intrinsic, AKA the Granite 3.2 8B Instruct
    Uncertainty LoRA. See model card [here](
        https://huggingface.co/ibm-granite/granite-uncertainty-3.2-8b-lora).
        
    The model must be prompted with a variant of the Granite 3.2 prompt. The model needs
    constrained decoding to produce a reliable result. The first token of the model
    output will be a certainty score encoded as a number from 0 to 9, inclusive.
    
    The processor as input a chat completion request and returns a completion with a 0.0
    to 0.9 certainty amount as a string in the content field.
    
    Example input to the IO processor's :func`acreate_chat_completion()` call: 
    
    ```
    {
        "messages": [
            {
                "role": "user", "content": "Can I go tubing along the Russian River in
                Sonoma County?"
            }
            {
                "role": "assistant", "content": "Yes, tubing along the Russian River in
                Sonoma County is a "
                           "popular activity. The recommended route for tubing is the "
                           "2-mile stretch from Steelhead Beach to Sunset Beach in "
                           "Forestville."
            },
        ], "documents": [
            {
                "text": "Sonoma County has many vacation destinations..."
            }, {
                "text": "Marin County also has vacation destinations..."
            }
        ], "generate_inputs": {
            "temperature": 0.0, "max_tokens": 4096,
        }
    }
    ```
    
    Example prompt that the IO processor would send to the model if it received the
    above input: ``` <|start_of_role|>documents<|end_of_role|> ... <|end_of_text|>
    <|start_of_role|>user<|end_of_role|>Can I go tubing along the Russian River in \
    Sonoma County?<|end_of_text|> <|start_of_role|>assistant<|end_of_role|>Yes, tubing
    along the Russian River in Sonoma County is a popular activity. The recommended
    route for tubing is the 2-mile stretch from Steelhead Beach to Sunset Beach in
    Forestville.<|end_of_text|> <|start_of_role|>certainty<|end_of_role|>```
    
    Example of raw output of the model for the above request: 
    ```
    7<|end_of_text|>
    ```
    (note that constrained decoding is required to produce the `<|end_of_text|>` token 
    reliably)
    
    Example of processed output from this IO processor for the above raw model output:
    ```
    {
        "results": [
            {
                "next_message": {
                    "role": "assistant",
                    "content": "0.7"
                }
            }
        ]
    }
    ```
    """

    def __init__(self, backend):
        super().__init__(backend=backend)

        # I/O processor for the base model, which does most of the input formatting.
        self.base_input_processor = Granite3Point2InputProcessor()

    def inputs_to_generate_inputs(
        self, inputs: ChatCompletionInputs, add_generation_prompt: bool = True
    ) -> GenerateInputs:
        # Validate the input and convert to Granite input
        inputs = Granite3Point2Inputs.model_validate(inputs.model_dump())

        # Check for the invariants that the model expects its input to satisfy
        if inputs.messages[-1].role not in ("user", "assistant"):
            raise ValueError("Last message is not a user or assistant message")

        # The beginning of the prompt doesn't change relative to base Granite 3.2
        prompt = self.base_input_processor.transform(inputs, False)

        # Only the generation prompt portion changes
        if add_generation_prompt:
            prompt = prompt + "<|start_of_role|>certainty<|end_of_role|>"
        result = inputs.generate_inputs.model_copy(
            update={
                "prompt": prompt,
                # Enable constrained decoding on vLLM backends
                "extra_body": {"guided_choice": [str(i) for i in range(10)]},
            }
        )
        return result

    def output_to_result(
        self, output: GenerateResults, inputs: ChatCompletionInputs | None = None
    ) -> ChatCompletionResults:
        results = []
        for raw_result in output.results:
            # Raw output should be a number from 0-9
            raw_str = raw_result.completion_string

            # Ignore everything after first character
            raw_number = raw_str[0]
            if raw_number.isdigit():
                parsed_prob = 0.1 * int(raw_number)
            else:
                parsed_prob = float("nan")  # Return nan on parse failure
            results.append(
                ChatCompletionResult(
                    next_message=AssistantMessage(
                        # Round to 1 digit so that we don't get 0.7 => 0.70000000001
                        content=f"{parsed_prob:0.1f}",
                        raw=raw_str,
                    )
                )
            )

        return ChatCompletionResults(results=results)


DEFAULT_CANNED_RESPONSE = (
    "Sorry, but I am unable to generate a response with high certainty."
)


class AssistantMessageWithScore(AssistantMessage):
    """Extended output format for the :class:`CertaintyCompositeIOProcessor` with
    an extra field for passing through certainty score."""

    certainty_score: float | None = None
    """Output of checking this message with the certainty intrinsic."""


class CertaintyCompositeIOProcessor(InputOutputProcessor):
    """
    Composite I/O processor that generates a response, checks the response for
    certainty, and falls back on a canned response if the check falls below a threshold.
    """

    def __init__(
        self,
        generator: InputOutputProcessor,
        lora_backend: Backend,
        threshold: float = 0.75,
        canned_response: str = DEFAULT_CANNED_RESPONSE,
        include_score: bool = False,
    ):
        """
        :param generator: I/O processor that generates the results that this I/O
         processor shoid validate.
        :param lora_backend: Backend for running the certainty intrinsic.
        :param threshold: If the certainty score of every completion is below this
         value, a canned response will be given.
        :param canned_response: Fallback response if none of the responses from the
         generator pass the threshold.
        """
        self._generator = generator
        self._certainty = CertaintyIOProcessor(lora_backend)
        self._threshold = threshold
        self._canned_response = canned_response
        self._include_score = include_score

    def update_threshold(self, threshold: float):
        """Convenience method to update the filtering threshold.

        :param threshold: New value to be applied to subsequent calls to the certainty
         I/O processor."""
        self._threshold = threshold

    async def acreate_chat_completion(
        self, inputs: ChatCompletionInputs
    ) -> ChatCompletionResults:
        # Generate one or more completions
        generator_output = await self._generator.acreate_chat_completion(inputs)

        # Run certainty checks on all completions in parallel
        futures = []
        for result in generator_output.results:
            futures.append(
                self._certainty.acreate_chat_completion(
                    inputs.with_next_message(
                        result.next_message
                    ).with_addl_generate_params({"n": 1, "temperature": 0.0})
                )
            )

        # Process results as they come back. Check each certainty score against the
        # threshold.
        processed_results = []
        max_certainty = 0.0
        for result, future in zip(generator_output.results, futures, strict=True):
            certainty_output = await future
            certainty_score = float(certainty_output.results[0].next_message.content)
            max_certainty = max(certainty_score, max_certainty)
            if certainty_score >= self._threshold:
                # Tack a certainty score onto the assistant message.
                message_with_score = AssistantMessageWithScore.model_validate(
                    result.next_message.model_dump()
                    | {"certainty_score": certainty_score}
                )
                processed_results.append(
                    result.model_copy(update={"next_message": message_with_score})
                )

        if len(processed_results) == 0:
            # No completions passed the certainty threshold. Use canned response.
            processed_results.append(
                ChatCompletionResult(
                    next_message=AssistantMessageWithScore(
                        content=self._canned_response, certainty_score=max_certainty
                    )
                )
            )

        return ChatCompletionResults(results=processed_results)



# SPDX-License-Identifier: Apache-2.0

"""
Test cases for the retrieval intrinsic.
"""

# Standard
import pathlib
import tempfile

# Third Party
import pytest
import torch

# Local
from granite_io.io.granite_3_2.input_processors.granite_3_2_input_processor import (
    Granite3Point2Inputs,
)
from granite_io.io.retrieval import (
    InMemoryRetriever,
    RetrievalRequestProcessor,
    compute_embeddings,
    write_embeddings,
)
from granite_io.io.retrieval.util import (
    read_mtrag_corpus,
)

_EXAMPLE_CHAT_INPUT = Granite3Point2Inputs.model_validate(
    {
        "messages": [
            {
                "role": "assistant",
                "content": "Welcome to the California Appellate Courts help desk.",
            },
            {
                "role": "user",
                "content": "I need to do some legal research to be prepared for my "
                "oral argument. Can I visit the law library?",
            },
        ],
        "generate_inputs": {
            "temperature": 0.0,
            "max_tokens": 4096,
        },
    }
)

_DATA_DIR = pathlib.Path("data/test_retrieval")
_EMBEDDING_MODEL_NAME = "multi-qa-mpnet-base-dot-v1"


@pytest.fixture
def govt_embeddings_file():
    """
    :returns: a pre-indexed copy of a tiny slice of the MTRAG benchmark's "govt" data
     set.
    """
    target_file = _DATA_DIR / "govt10_embeds.parquet"
    return target_file


@pytest.fixture
def govt_docs_file():
    """
    :returns: a copy of a tiny slice of the MTRAG benchmark's "govt" data set.
    """
    target_file = _DATA_DIR / "govt10.jsonl.zip"
    return target_file


def test_make_embeddings(govt_docs_file):  # pylint: disable=redefined-outer-name
    """
    Verify that embedding creation is working by creating embeddings for the first
    10 documents in the "govt" corpus from the MTRAG benchmark.
    """
    full_govt = read_mtrag_corpus(govt_docs_file)
    corpus = full_govt.slice(0, 10)
    embeddings = compute_embeddings(corpus, _EMBEDDING_MODEL_NAME)
    assert embeddings.column("embedding").to_pylist()[0][:10] == pytest.approx(
        [
            -0.11038033664226532,
            0.27693408727645874,
            -0.11863572895526886,
            -0.0792723074555397,
            0.20247098803520203,
            0.09491363912820816,
            0.6091732978820801,
            0.0905776172876358,
            0.10194987803697586,
            -0.011982650496065617,
        ],
        abs=1e-3,
    )

    # Round-trip through a file and make sure we get the embeddings back.
    with tempfile.TemporaryDirectory() as tmpdir:
        file_loc = write_embeddings(tmpdir, "test", embeddings)
        retriever = InMemoryRetriever(file_loc, _EMBEDDING_MODEL_NAME)
        # pylint: disable=protected-access
        assert retriever._embeddings[1] == pytest.approx(
            torch.tensor(embeddings.column("embedding").to_pylist()[1]), abs=1e-3
        )


def test_in_memory_retriever(govt_embeddings_file):  # pylint: disable=redefined-outer-name
    """Verify basic functionality of the InMemoryRetriever class"""
    retriever = InMemoryRetriever(govt_embeddings_file, _EMBEDDING_MODEL_NAME)
    result = retriever.retrieve(_EXAMPLE_CHAT_INPUT.messages[-1].content)
    assert result.column("id").to_pylist() == [
        "775449d1aa187ec5",
        "775449d1aa187ec5",
        "775449d1aa187ec5",
        "775449d1aa187ec5",
        "775449d1aa187ec5",
    ]


def test_retrieval_request_processor(govt_embeddings_file):  # pylint: disable=redefined-outer-name
    """Basic test of the RequestProcessor that performs RAG retrieval"""
    retriever = InMemoryRetriever(govt_embeddings_file, _EMBEDDING_MODEL_NAME)
    request_processor = RetrievalRequestProcessor(retriever, top_k=3)
    results = request_processor.process(_EXAMPLE_CHAT_INPUT)
    assert len(results) == 1
    # print(results[0].documents)
    assert [d.doc_id for d in results[0].documents] == [
        "775449d1aa187ec5",
        "775449d1aa187ec5",
        "775449d1aa187ec5",
    ]



# SPDX-License-Identifier: Apache-2.0

"""
Simple placeholder for unit testing contrib
"""

# Third Party
from src.granite_io_pdl.pdl_io import (  # pylint: disable=import-error
    PdlInputOutputProcessor,
)

pdl_processor = """
function:
  inputs: obj
return:
  lastOf:
  - model: "granite3.2:2b"
    backend: openai
    input: ${ inputs.messages }
    modelResponse: results
  - ${ results }
"""


def test_pdl_io():
    """Simple example test"""
    ret = PdlInputOutputProcessor(pdl=pdl_processor)
    assert ret is not None



# SPDX-License-Identifier: Apache-2.0

"""
The backend module provides abstractions and implementations for connecting to
models
"""

# Local
from granite_io.backend.base import Backend, ChatCompletionBackend  # noqa: F401
from granite_io.backend.registry import backend, make_backend  # noqa: F401
import granite_io.backend.litellm  # noqa: F401
import granite_io.backend.openai  # noqa: F401
import granite_io.backend.transformers  # noqa: F401
import granite_io.backend.vllm_server  # noqa: F401



# SPDX-License-Identifier: Apache-2.0

"""
The io module holds the abstraction and implementations for Granite 3.2 IO
processing.
"""

# Local
import granite_io.io.granite_3_2.granite_3_2  # noqa: F401



# SPDX-License-Identifier: Apache-2.0
# ruff: noqa: E501
# pylint: disable=redefined-outer-name,broad-exception-caught

# Standard
import json
import os

# Third Party
from litellm import UnsupportedParamsError
from openai import APIConnectionError, PermissionDeniedError
from pydantic import ValidationError
import pytest
import transformers

# Local
from granite_io import make_io_processor
from granite_io.backend import Backend, make_backend
from granite_io.backend.litellm import LiteLLMBackend
from granite_io.backend.openai import OpenAIBackend
from granite_io.backend.transformers import TransformersBackend
from granite_io.io.consts import (
    _GRANITE_3_2_2B_HF,
    _GRANITE_3_2_COT_END,
    _GRANITE_3_2_COT_START,
    _GRANITE_3_2_MODEL_NAME,
)
from granite_io.io.granite_3_2.granite_3_2 import (
    Granite3Point2InputOutputProcessor,
)
from granite_io.io.granite_3_2.input_processors.granite_3_2_input_processor import (
    ControlsRecord,
    Granite3Point2Inputs,
)
from granite_io.io.granite_3_2.output_processors.granite_3_2_output_parser import (
    _CITATION_START,
    _HALLUCINATION_START,
)
from granite_io.io.granite_3_2.output_processors.granite_3_2_output_processor import (
    _COT_END_ALTERNATIVES,
    _COT_START_ALTERNATIVES,
)
from granite_io.types import (
    AssistantMessage,
    ChatCompletionInputs,
    ChatCompletionResult,
    ChatCompletionResults,
    Citation,
    Document,
    GenerateInputs,
    GenerateResult,
    GenerateResults,
    Hallucination,
    UserMessage,
)

## Helpers #####################################################################

# All the different chat completion requests that are tested in this file, serialized as
# JSON strings. Represented as a dictionary instead of a list so that pytest output will
# show the short key instead of the long value when referencing a single run of a test
INPUT_JSON_STRS = {
    "simple": """
{
    "messages":
    [
        {"role": "user", "content": "Hello, how are you?"},
        {"role": "assistant", "content": "I'm doing great. How can I help you today?"},
        {"role": "user", "content": "I'd like to show off how chat templating works!"}
    ]
}
""",
    "thinking_tag": """
{
    "messages":
    [
        {"role": "user", "content": "How much wood could a wood chuck chuck?"}
    ],
    "thinking": true
}
""",
    "stop_strings": """
{
    "messages":
        [
            {"role": "user", "content": "How much wood could a wood chuck chuck?"}
        ],
    "generate_inputs":
        {
            "stop": "woodchuck"
        }
}
""",
    "custom_system_prompt": """
{
    "messages":
    [
        {"role": "system", "content": "Answer all questions like a three year old."},
        {"role": "user", "content": "Hi, I would like some advice on the best tax \
strategy for managing dividend income."}
    ]
}
""",
}


@pytest.fixture(scope="session", params=INPUT_JSON_STRS)
def input_json_str(request: pytest.FixtureRequest) -> str:
    """Pytest fixture that allows us to run a given test case repeatedly with multiple
    different chat completion requests."""
    return INPUT_JSON_STRS[request.param]


@pytest.fixture(scope="session")
def tokenizer() -> transformers.PreTrainedTokenizerBase:
    model_path = _GRANITE_3_2_2B_HF
    try:
        ret = transformers.AutoTokenizer.from_pretrained(
            model_path, local_files_only=False
        )
    except Exception as e:
        pytest.skip(f"No tokenizer for {model_path}: {e}")
    return ret


msg = UserMessage(content="Hello")
no_thinking_input = ChatCompletionInputs(messages=[msg])
thinking_input = ChatCompletionInputs(messages=[msg], thinking=True)

thought = "Think think"
response = "respond respond"
pre_thought = "something before"
no_cot_output = f"{thought} {response}"
no_thinking_output = f"{thought} {_GRANITE_3_2_COT_END} {response}"
no_response_output = f"{_GRANITE_3_2_COT_START}\n\n{response}"
cot_output = (
    f"{_GRANITE_3_2_COT_START}\n\n{thought}\n{_GRANITE_3_2_COT_END}\n\n{response}"
)
cot_alt_output = f"{_COT_START_ALTERNATIVES[-1]}\n\n{thought}\n{_COT_END_ALTERNATIVES[-1]}\n\n{response}"
cot_mixed_output = (
    f"{_GRANITE_3_2_COT_START}\n\n{thought}\n{_COT_END_ALTERNATIVES[-1]}\n\n{response}"
)
cot_pre_output = f"{pre_thought} {_GRANITE_3_2_COT_START} {thought} {_COT_END_ALTERNATIVES[-1]} {response}"

no_constituent_output = "Mad about dog!"
citation_example = '<co>1</co> Document 0: "Dog info"'
citation_output = (
    f"{no_constituent_output}<co>1</co>\n\n{_CITATION_START}\n\n{citation_example}\n\n"
)
hallucination_example = "1. Risk low: Mad about dog"
citation_hallucination_output = (
    f"{citation_output}{_HALLUCINATION_START}\n\n{hallucination_example}\n\n"
)
expected_citation = Citation(
    citation_id="1",
    doc_id="0",
    context_text="Dog info",
    context_begin=0,
    context_end=8,
    response_text="Mad about dog!",
    response_begin=0,
    response_end=14,
)
expected_document = Document(doc_id="0", text="Dog info")
doc_input = ChatCompletionInputs(
    messages=[msg], documents=[{"doc_id": "0", "text": "Dog info"}]
)
expected_hallucination = Hallucination(
    hallucination_id="1",
    risk="low",
    response_text="Mad about dog",
    response_begin=0,
    response_end=13,
)
## Tests #######################################################################


@pytest.mark.parametrize(
    ["length", "originality", "error"],
    [
        (None, None, None),
        ("short", None, None),
        (None, "abstractive", None),
        ("long", "extractive", None),
        ("BAD_VAL", "abstractive", "input_value='BAD_VAL'"),
        ("long", "BAD_VAL", "input_value='BAD_VAL'"),
        ("BAD_VAL", "Another Bad Value", "input_value='BAD_VAL'"),
        ("ShOrT", None, "input_value='ShOrT'"),
        (None, "aBsTrAcTiVe", "input_value='aBsTrAcTiVe'"),
        (1, None, "input_type=int"),
        (None, 2, "input_type=int"),
    ],
)
def test_controls_field_validators(length, originality, error):
    if error:
        with pytest.raises(ValidationError, match=error):
            ControlsRecord(length=length, originality=originality)
    else:
        ControlsRecord(length=length, originality=originality)


def test_read_inputs(input_json_str):
    """
    Verify that the dataclasses for the Granite 3.2 I/O processor can parse Granite 3.2 JSON
    """
    input_json = json.loads(input_json_str)
    input_obj = ChatCompletionInputs.model_validate(input_json)
    input_obj_2 = ChatCompletionInputs.model_validate_json(input_json_str)

    assert input_obj == input_obj_2

    # Parse additional Granite-specific fields
    granite_input_obj = Granite3Point2Inputs.model_validate(input_obj.model_dump())

    # Verify that we can convert back to JSON without crashing
    granite_input_obj.model_dump_json()
    input_obj.model_dump_json()


def test_same_input_string(
    tokenizer: transformers.PreTrainedTokenizerBase, input_json_str: str
):
    """
    Verify that the I/O processor produces the exact same input string as the Jinja
    template that ships with the model.
    """

    # First apply the Jinja template
    input_json = json.loads(input_json_str)
    input_kwargs = input_json.copy()
    del input_kwargs["messages"]
    transformers_str = tokenizer.apply_chat_template(
        input_json["messages"],
        **input_kwargs,
        tokenize=False,
        add_generation_prompt=True,
    )

    # Then compare against the input processor
    inputs = ChatCompletionInputs.model_validate_json(input_json_str)
    io_proc_str = Granite3Point2InputOutputProcessor().inputs_to_string(inputs)

    assert io_proc_str == transformers_str


def test_basic_inputs_to_string():
    """
    Basic test against canned output in case the developer doesn't have a way to load
    an actual Granite 3.2 tokenizer for output comparisons.

    Chat input:

    chat = [
        {"role": "user", "content": "Hello, how are you?"},
        {"role": "assistant", "content": "I'm doing great. How can I help you today?"},
        {"role": "user", "content": "I'd like to show off how chat templating works!"},
    ]

    Expected similar (dates will vary) chat template request generated:

    <|start_of_role|>system<|end_of_role|>Knowledge Cutoff Date: April 2024.
    Today's Date: February 17, 2025.
    You are Granite, developed by IBM. You are a helpful AI assistant.<|end_of_text|>
    <|start_of_role|>user<|end_of_role|>Hello, how are you?<|end_of_text|>
    <|start_of_role|>assistant<|end_of_role|>I'm doing great. How can I help you today?<|end_of_text|>
    <|start_of_role|>user<|end_of_role|>I'd like to show off how chat templating works!<|end_of_text|>
    """
    chatRequest = Granite3Point2InputOutputProcessor().inputs_to_string(
        inputs=ChatCompletionInputs(
            messages=[
                UserMessage(content="Hello, how are you?"),
                AssistantMessage(content="I'm doing great. How can I help you today?"),
                UserMessage(content="I'd like to show off how chat templating works!"),
            ]
        ),
        add_generation_prompt=False,
    )

    chatReqStart = "<|start_of_role|>system<|end_of_role|>Knowledge Cutoff Date:"
    assert chatRequest.startswith(chatReqStart)

    chatReqModelMsg = "You are Granite, developed by IBM. You are a helpful AI assistant.<|end_of_text|>"
    assert chatReqModelMsg in chatRequest

    chatReqBody = """\
<|start_of_role|>user<|end_of_role|>Hello, how are you?<|end_of_text|>
<|start_of_role|>assistant<|end_of_role|>I'm doing great. How can I help you today?<|end_of_text|>
<|start_of_role|>user<|end_of_role|>I'd like to show off how chat templating works!<|end_of_text|>"""
    assert chatReqBody in chatRequest

    assert chatRequest.endswith("")


@pytest.mark.vcr
def test_completion_repetition_param(backend_x: Backend):
    messages = [
        {
            "role": "user",
            "content": "Can you answer my question?",
        }
    ]

    # What a client might be sending to a backend
    generate_inputs = {
        "prompt": "Just give me an example of what you can do",
        "temperature": 0.5,
        "repetition_penalty": 0.5,
    }
    inputs = ChatCompletionInputs(messages=messages, generate_inputs=generate_inputs)

    io_processor = make_io_processor(_GRANITE_3_2_MODEL_NAME, backend=backend_x)
    try:
        outputs: ChatCompletionResults = io_processor.create_chat_completion(inputs)
    except TypeError as te:
        if isinstance(backend_x, OpenAIBackend):
            pytest.xfail(str(te))
        raise te
    except UserWarning as uw:
        if isinstance(backend_x, TransformersBackend):
            pytest.xfail(str(uw))
        raise uw

    assert isinstance(outputs, ChatCompletionResults)


@pytest.mark.vcr
def test_completion_presence_param(backend_x: Backend):
    messages = [
        {
            "role": "user",
            "content": "Can you answer my question?",
        }
    ]
    generate_inputs = {
        "prompt": "Just give me an example of what you can do",
        "temperature": 0.5,
        "presence_penalty": 0.5,
        "frequency_penalty": 0.5,
    }
    inputs = ChatCompletionInputs(messages=messages, generate_inputs=generate_inputs)

    io_processor = make_io_processor(_GRANITE_3_2_MODEL_NAME, backend=backend_x)
    try:
        outputs: ChatCompletionResults = io_processor.create_chat_completion(inputs)
    except UnsupportedParamsError as upe:
        # Specific exception from LiteLLMBackend
        pytest.xfail(upe.message)
    except UserWarning as uw:
        if isinstance(backend_x, TransformersBackend):
            pytest.xfail(str(uw))
        raise uw

    assert isinstance(outputs, ChatCompletionResults)


@pytest.mark.vcr(record_mode="new_episodes")
@pytest.mark.vcr
def test_run_processor(backend_x: Backend, input_json_str: str):
    inputs = ChatCompletionInputs.model_validate_json(input_json_str)
    io_processor = make_io_processor(_GRANITE_3_2_MODEL_NAME, backend=backend_x)
    outputs: ChatCompletionResults = io_processor.create_chat_completion(inputs)

    assert isinstance(outputs, ChatCompletionResults)
    assert len(outputs.results) == 1

    content = outputs.results[0].next_message.content
    assert content  # Make sure we don't get empty result (I had a bug)

    # Test for stop reason
    if inputs.generate_inputs and inputs.generate_inputs.stop:
        stop = inputs.generate_inputs.stop
        # Note: currently transformers includes the stop tokens,
        # but OpenAI does not. So, either endswith or not-in.
        assert stop not in content
        assert outputs.results[0].next_message.stop_reason == "stop"

    # TODO: Verify outputs in greater detail


@pytest.mark.parametrize(
    ["inputs", "output", "exp_thought", "exp_resp"],
    [
        # No thinking flag
        (no_thinking_input, no_thinking_output, None, no_thinking_output),
        (no_thinking_input, cot_output, None, cot_output),
        # Thinking flag
        (thinking_input, no_cot_output, None, no_cot_output),
        (thinking_input, no_thinking_output, None, no_thinking_output),
        (thinking_input, no_response_output, None, no_response_output),
        (thinking_input, cot_output, thought, response),
        (thinking_input, cot_alt_output, thought, response),
        (thinking_input, cot_mixed_output, thought, response),
        (thinking_input, cot_pre_output, thought, f"{pre_thought} {response}"),
    ],
)
def test_cot_parsing(inputs, output, exp_thought, exp_resp):
    """Test the parsing logic for CoT reasoning output"""
    proc = Granite3Point2InputOutputProcessor()
    generated = GenerateResults(
        results=[
            GenerateResult(
                completion_string=output, completion_tokens=[], stop_reason="?"
            )
        ]
    )
    result = proc.output_to_result(generated, inputs).results[0].next_message
    assert result.reasoning_content == exp_thought
    assert result.content == exp_resp
    assert result.raw == output


@pytest.mark.parametrize(
    [
        "inputs",
        "output",
        "exp_document",
        "exp_citation",
        "exp_hallucination",
        "exp_resp",
    ],
    [
        # No constituents
        (
            no_thinking_input,
            no_constituent_output,
            None,
            None,
            None,
            no_constituent_output,
        ),
        # Citation
        (
            doc_input,
            citation_output,
            [expected_document],
            [expected_citation],
            None,
            no_constituent_output,
        ),
        # Citation and hallucination
        (
            doc_input,
            citation_hallucination_output,
            [expected_document],
            [expected_citation],
            [expected_hallucination],
            no_constituent_output,
        ),
    ],
)
def test_citation_hallucination_parsing(
    inputs, output, exp_document, exp_citation, exp_hallucination, exp_resp
):
    """Test the parsing logic for Rag and hallucinations output"""
    proc = Granite3Point2InputOutputProcessor()
    generated = GenerateResults(
        results=[
            GenerateResult(
                completion_string=output, completion_tokens=[], stop_reason="?"
            )
        ]
    )
    result = proc.output_to_result(generated, inputs).results[0].next_message
    assert result.content == exp_resp
    assert result.citations == exp_citation
    assert result.documents == exp_document
    assert result.hallucinations == exp_hallucination


@pytest.mark.vcr(record_mode="new_episodes")
@pytest.mark.block_network
def test_multiple_return(backend_x: Backend, input_json_str: str):
    inputs = ChatCompletionInputs.model_validate_json(input_json_str)
    inputs = inputs.model_copy(
        update={"generate_inputs": GenerateInputs(max_tokens=1024, n=3)}
    )
    io_processor = make_io_processor(_GRANITE_3_2_MODEL_NAME, backend=backend_x)
    try:
        results: ChatCompletionResults = io_processor.create_chat_completion(inputs)
    except UnsupportedParamsError:
        if isinstance(backend_x, LiteLLMBackend):
            pytest.xfail("LiteLLMBackend support for n > 1 varies by provider")

    assert isinstance(results, ChatCompletionResults)

    # OpenAI backend with custom system prompt returns 1 result only
    assert len(results.results) == 1 or len(results.results) == 3

    # TODO: Verify outputs in greater detail


test_key = "RITS_API_KEY"
open_api_key = os.environ.get("OPENAI_API_KEY", "notset")
bogus_api_key = "this-is-a-bogus-api-key"
test_header = {test_key: open_api_key}
bogus_header = {test_key: bogus_api_key}


def header_id(default_headers, extra_headers):
    dh_value = default_headers.get(test_key) if default_headers else None
    eh_value = extra_headers.get(test_key) if extra_headers else None

    ids = {
        open_api_key: "notset" if open_api_key == "notset" else "env",
        bogus_api_key: "bogus",
        None: "none",
    }
    return f"{ids[dh_value]}-{ids[eh_value]}"


header_test_matrix = [
    (default_headers, extra_headers)
    for default_headers in (None, bogus_header, test_header)
    for extra_headers in (None, bogus_header, test_header)
]
header_test_ids = [header_id(h[0], h[1]) for h in header_test_matrix]


@pytest.mark.parametrize(
    "default_headers, extra_headers", header_test_matrix, ids=header_test_ids
)
@pytest.mark.skipif(open_api_key == "notset", reason="Needs OpenAI API key")
async def test_headers(request, default_headers, extra_headers):
    """Test headers using manual test case against a server to verify.

    Use environment variables suitable for an OpenAI backend:
    OPENAI_BASE_URI - base URL to the provider
    MODEL_NAME - model name override for the provider
    OPENAI_API_KEY - api key used for the provider
    Note:  For this specific test, we have a provider that needs
           the API key to also be in an additional header.
           It also passes the num_returned test (Ollama does not).

    Testing default_headers (openai only) in the openai contructor.
    Testing extra_headers in the openai or litellm call.
    Testing with both or neither.
    """
    be = make_backend(
        "openai",
        {
            "model_name": "ibm-granite/granite-8b-instruct-preview-4k",
            "openai_api_key": "<your api key>",  # Use env var to set
            "openai_base_url": "<your base url>",  # Use env var to set
            "default_headers": default_headers,
        },
    )

    inputs = ChatCompletionInputs(
        messages=[
            {
                "role": "user",
                "content": "what is up?",
            }
        ],
        generate_inputs={
            "extra_headers": extra_headers,
            "n": 3,  # n sampling works on test backend
        },
    )

    io_processor = make_io_processor(_GRANITE_3_2_MODEL_NAME, backend=be)

    test_name = request.node.name
    if "env" in test_name and "env-bogus" not in test_name:
        # when env is set unless a bogus extra_header overrides default_header
        output: ChatCompletionResults = io_processor.create_chat_completion(inputs)
        results = output.results
        for result in results:
            assert isinstance(result, ChatCompletionResult)
        assert len(results) == 3  # Note: This would fail if you just hit Ollama
    else:
        # Raises with either bad apikey or bad url (for testing bogus, none and notset)
        with pytest.raises((APIConnectionError, PermissionDeniedError)):
            io_processor.create_chat_completion(inputs)



# SPDX-License-Identifier: Apache-2.0

"""
I/O processor for the Granite hallucinations intrinsic.
"""

# Standard
import json

# Third Party
import nltk

# Local
from granite_io.backend.base import Backend
from granite_io.io.base import (
    InputOutputProcessor,
    ModelDirectInputOutputProcessorWithGenerate,
)
from granite_io.io.granite_3_2.input_processors.granite_3_2_input_processor import (
    ControlsRecord,
    Granite3Point2InputProcessor,
    Granite3Point2Inputs,
)
from granite_io.types import (
    ChatCompletionInputs,
    ChatCompletionResult,
    ChatCompletionResults,
    GenerateInputs,
    GenerateResults,
    Hallucination,
)

# The hallucinations intrinsic model expects to see a special system prompt after the
# last assistant message.
_HALLUCINATION_SYSTEM_PROMPT = (
    "Split the last assistant response into individual sentences. "
    "For each sentence in the last assistant response, identify the faithfulness "
    "score range. "
    "Ensure that your output includes all response sentence IDs, and for each response "
    "sentence ID, provide the corresponding faithfulness score range. "
    "The output must be a json structure."
)

_FAITHFULNESS_RISK_CONVERT = {
    "0.1-0": "0.9-1.0",
    "0.1-0.2": "0.8-0.9",
    "0.2-0.3": "0.7-0.8",
    "0.3-0.4": "0.6-0.7",
    "0.4-0.5": "0.5-0.6",
    "0.5-0.6": "0.4-0.5",
    "0.6-0.7": "0.3-0.4",
    "0.7-0.8": "0.2-0.3",
    "0.8-0.9": "0.1-0.2",
    "0.9-1.0": "0-0.1",
    "unanswerable": "unanswerable",
    "NA": "NA",
}


def faithfulness_to_risk(input_string):
    """
    Takes an input string returns the corresponding
    output string from the dictionary. Uses try-except for error handling.
    """
    try:
        return _FAITHFULNESS_RISK_CONVERT[input_string]
    except KeyError:
        # Handle the case where the key is not found
        return "NA"  # Or raise an exception, or return a default value


def mark_sentence_boundaries(
    split_strings: list[list[str]], tag_prefix: str
) -> tuple[str, int]:
    """
    Modify one or more input strings by inserting a tag in the form
    ``<[prefix][number]>``
    at the location of each sentence boundary.

    :param split_strings: Input string(s), pre-split into sentences
    :param tag_prefix: String to place before the number part of each tagged
        sentence boundary.

    :returns: List of input strings with all sentence boundaries marked.
    """
    index = 0
    result = []
    for sentences in split_strings:
        to_concat = []
        for sentence in sentences:
            to_concat.append(f"<{tag_prefix}{index}> {sentence}")
            index += 1
        result.append(" ".join(to_concat))
    return result


class HallucinationsIOProcessor(ModelDirectInputOutputProcessorWithGenerate):
    """
    I/O processor for the Granite hallucinations intrinsic, also known as the 
    [LoRA Adapter for Hallucination Detection in RAG outputs](
        https://huggingface.co/ibm-granite/
        granite-3.2-8b-lora-rag-hallucination-detection
    )

    Example input to the IO processor's :func`acreate_chat_completion()` call:

    ```
    {
    "messages": [
        {
            "role": "user",
            "content": "What is the visibility level of Git Repos and Issue Tracking \
projects?"
        },
        {
            "role": "assistant",
            "content": "Git Repos and Issue Tracking projects can have one of three \
visibility levels: private, internal, or public. Private projects are visible only to \
project members, internal projects are visible to all users logged in to IBM Cloud, \
and public projects are visible to anyone."
        }
    ],
    "documents": [
        {
            "text": "Git Repos and Issue Tracking is an IBM-hosted component of the \
Continuous Delivery service. All of the data that you provide to Git Repos and Issue \
Tracking, including but not limited to source files, issues, pull requests, and \
project configuration properties, is managed securely within Continuous Delivery. \
However, Git Repos..."
        },
        {
            "text": "After you create a project in Git Repos and Issue Tracking, but \
before you entrust any files, issues, records, or other data with the project, review \
the project settings and change any settings that are necessary to protect your data. \
Settings to review include visibility levels, email notifications, integrations, web \
hooks, access tokens, deploy tokens, and deploy keys...."
        }
    ],
    "generate_inputs": {
        "temperature": 0.0,
        "max_tokens": 1024
    }
}
    ```

    Example prompt that the IO processor would send to the model if it received the
    above input:

    ```
<|start_of_role|>system<|end_of_role|>Knowledge Cutoff Date: April 2024.
Today's Date: April 09, 2025.
You are Granite, developed by IBM. You are a helpful AI assistant with access to the \
following tools. When a tool is required to answer the user's query, respond with \
<|tool_call|> followed by a JSON list of tools used. If a tool does not exist in the \
provided list of tools, notify the user that you do not have the ability to fulfill \
the request.<|end_of_text|>
<|start_of_role|>documents<|end_of_role|>Document 0
Git Repos and Issue Tracking is an IBM-hosted component of the Continuous \
Delivery service. All of the data that you provide to Git Repos and Issue \
Tracking, including but not limited to source files, issues, pull requests, and \
project configuration properties, is managed securely within Continuous Delivery. 
[...]

Document 1
After you create a project in Git Repos and Issue Tracking, but before you \
entrust any files, issues, records, or other data with the project, review the \
project settings and change any settings that are necessary to protect your data. \
Settings to review include visibility levels, email notifications, integrations, \
web hooks, access tokens, deploy tokens, and deploy keys. Project visibility \
[...]

<|start_of_role|>user<|end_of_role|>What is the visibility level of Git Repos and \
Issue Tracking projects?<|end_of_text|>
<|start_of_role|>assistant<|end_of_role|><r0> Git Repos and Issue Tracking projects \
can have one of three visibility levels: private, internal, or public. <r1> Private \
projects are visible only to project members, internal projects are visible to all \
users logged in to IBM Cloud, and public projects are visible to anyone.\
<|end_of_text|> \
<|start_of_role|>system<|end_of_role|>"Split the last assistant response into \
individual sentences. \
For each sentence in the last assistant response, identify the faithfulness score \
range. \
Ensure that your output includes all response sentence IDs, and for each response \
sentence ID, \
provide the corresponding faithfulness score range. \
The output must be a json structure.<|end_of_text|>
    ```

    Example of raw output of the model for the above request:

    ```
    "{\\"<r0>\\": \\"0.2-0.3\\", \\"<r1>\\": \\"0.9-1.0\\"}"
    ```
    Note that the raw model output is JSON data encoded as a JSON string.

    Example of processed output from this IO processor for the above raw model output:
    ```
    {
    "content": "Git Repos and Issue Tracking projects can have one of three visibility \
levels: private, internal, or public. Private projects are visible only to \
project members, internal projects are visible to all users logged in to IBM Cloud, \
and public projects are visible to anyone.",
    "role": "assistant",
    "tool_calls": [],
    "reasoning_content": null,
    "hallucinations": [
        {
          "hallucination_id": "0",
          "risk": "0.1-0.2",
          "response_text": "Git Repos and Issue Tracking projects can have one of \
three visibility levels: private, internal, or public.",
          "response_begin": 0,
          "response_end": 108
        },
        {
          "hallucination_id": "1",
          "risk": "0.9-1.0",
          "response_text": "Private projects are only visible to project members, \
internal projects are visible to all users logged in to IBM Cloud, and public projects \
are visible to anyone.",
          "response_begin": 109,
          "response_end": 272
        }
    ],
    "hallucinations": null,
    "hallucinations": null,
    "stop_reason": null
    }
    ```
    """

    def __init__(self, backend):
        super().__init__(backend=backend)

        # Input processor for the base model, which does most of the input formatting.
        self.base_input_processor = Granite3Point2InputProcessor()

        # Object that identifies sentence boundaries. Currently we assume an NLTK
        # sentence tokenizer is used here. This may change in the future.
        # Note that THE SENTENCE SPLITTER MUST BE DETERMINISTIC, because we invoke it
        # twice on every inference request.
        self.sentence_splitter = nltk.tokenize.punkt.PunktSentenceTokenizer()

    def inputs_to_generate_inputs(
        self, inputs: ChatCompletionInputs, add_generation_prompt: bool = True
    ) -> GenerateInputs:
        # Validate the input and convert to Granite input
        inputs = Granite3Point2Inputs.model_validate(inputs.model_dump())

        # Check for the invariants that the model expects its input to satisfy
        if not inputs.messages[-1].role == "assistant":
            raise ValueError("Last message is not an assistant message")
        if inputs.documents is None or len(inputs.documents) == 0:
            raise ValueError("Input does not contain documents")

        # The last assistant message also needs to be split into sentences. The we
        # encode it with each sentence boundary marked with <r0>, <r1>, ... <rj-1>,
        # where `j` is the number of sentences in the message.
        last_message_as_sentences = list(
            self.sentence_splitter.tokenize(inputs.messages[-1].content)
        )

        rewritten_last_message_text = mark_sentence_boundaries(
            [last_message_as_sentences], "r"
        )[0]
        rewritten_messages = [m.model_copy() for m in inputs.messages]
        rewritten_messages[-1].content = rewritten_last_message_text

        # Put the rewritten docs and last message back into the original chat completion
        # and let the Granite 3.2 IO processor take care of the rest of the formatting.
        rewritten_inputs = inputs.model_copy(
            update={"documents": inputs.documents, "messages": rewritten_messages}
        )
        prompt = self.base_input_processor.transform(
            rewritten_inputs,
            # No <|start_of_role|>assistant<|end_of_role|> at end of prompt string.
            False,
        )

        # The model's training data uses a special system message to the end of the
        # messages as the prompting string.
        # This is not a valid input for the base model, so we append it to the prompt
        # manually.
        if add_generation_prompt:
            prompt = prompt + (
                f"<|start_of_role|>system<|end_of_role|>{_HALLUCINATION_SYSTEM_PROMPT}"
                f"<|end_of_text|>"
            )

        generate_inputs_before = (
            inputs.generate_inputs if inputs.generate_inputs else GenerateInputs()
        )

        result = generate_inputs_before.model_copy(
            update={
                "prompt": prompt,
                # Single completion for hallucinations intrinsic
                "n": 1,
                # Always generate hallucinations at temperature 0
                "temperature": 0.0,
                # Ensure we have enough of a token budget to reliably produce the
                # full output.
                "max_tokens": 1024,
            }
        )
        return result

    def output_to_result(
        self, output: GenerateResults, inputs: ChatCompletionInputs | None = None
    ) -> ChatCompletionResults:
        if inputs is None:
            raise ValueError("Cannot construct results without original inputs")

        # Do the same for the original response
        message_sentence_offsets = list(
            self.sentence_splitter.span_tokenize(inputs.messages[-1].content)
        )

        results = []
        for raw_result in output.results:
            try:
                # Output of the model is JSON packed into a quoted string for some
                # reason. So parse twice.
                parsed_json = json.loads(raw_result.completion_string)
                parsed_json = json.loads(parsed_json)
                hallucinations = []
                next_hallucination_id = 0
                content = inputs.messages[-1].content

                if not isinstance(parsed_json, dict):
                    raise TypeError(
                        f"Model output '{raw_result.completion_string}' "
                        f"is not a JSON object"
                    )
                for response_key, value in parsed_json.items():
                    # Example: <r0>
                    response_index = int(response_key[2:-1])
                    response_begin, response_end = message_sentence_offsets[
                        response_index
                    ]
                    response_text = content[response_begin:response_end]
                    if not isinstance(value, str):
                        raise TypeError(f"Entry for {response_key} is not a str")
                    hallucinations.append(
                        Hallucination(
                            hallucination_id=str(next_hallucination_id),
                            response_text=response_text,
                            risk=faithfulness_to_risk(value),
                            response_begin=response_begin,
                            response_end=response_end,
                        )
                    )
                    next_hallucination_id += 1  # noqa: SIM113

            except (json.JSONDecodeError, TypeError) as e:
                # This type of error shouldn't happen, because we use constrained
                # decoding.
                content = f"ERROR: {e} (raw output: {raw_result})"
                hallucinations = []

                # TEMPORARY: Pass through errors for now
                raise e

            # print(f"Adding {raw_result.completion_string} as raw result")
            next_message = inputs.messages[-1].model_copy(
                update={
                    "content": content,
                    "hallucinations": hallucinations,
                    # TEMPORARY -- should be original message's raw result
                    "raw": raw_result.completion_string,
                }
            )

            results.append(ChatCompletionResult(next_message=next_message))

        return ChatCompletionResults(results=results)


class HallucinationsCompositeIOProcessor(InputOutputProcessor):
    """
    Composite I/O processor that generates a response, then adds hallucinations.
    """

    def __init__(
        self,
        generator: InputOutputProcessor,
        lora_backend: Backend,
        request_hallucinations_from_generator: bool = False,
    ):
        """
        :param generator: I/O processor that generates the results that this I/O
         processor shoid validate.
        :param lora_backend: Backend for running the hallucinations intrinsic.
        :param request_hallucinations_from_generator: if ``True``, invoke ``generator``
         with the Granite ``hallucinations`` control turned on. If ``False``, the
         value of the ``hallucinations`` control will be passed through from requests
         as is.
        """
        self._generator = generator
        self._hallucinations = HallucinationsIOProcessor(lora_backend)
        self._request_hallucinations_from_generator = (
            request_hallucinations_from_generator
        )

    def update_request_hallucinations_from_generator(
        self, request_hallucinations_from_generator: bool
    ):
        """Convenience method to update whether to request (and then discard)
        hallucinations output from the generator

        :param request_hallucinations_from_generator: New value to be applied to
         subsequent calls to the I/O processor."""
        self.request_hallucinations_from_generator = (
            request_hallucinations_from_generator
        )

    async def acreate_chat_completion(
        self, inputs: ChatCompletionInputs
    ) -> ChatCompletionResults:
        # Downcast to extended Granite inputs. This also creates a copy.
        inputs = Granite3Point2Inputs.model_validate(inputs.model_dump())

        if self._request_hallucinations_from_generator:
            # Code above already copied inputs, so we can modify inputs in place
            controls = ControlsRecord() if not inputs.controls else inputs.controls
            controls.hallucinations = True
            inputs.controls = controls

        generator_output = await self._generator.acreate_chat_completion(inputs)

        # Generate hallucinations for all completions in parallel
        futures = []
        for result in generator_output.results:
            futures.append(
                self._hallucinations.acreate_chat_completion(
                    inputs.with_next_message(
                        result.next_message
                    ).with_addl_generate_params(
                        # Only request the top-1 result from the LoRA
                        {"n": 1, "temperature": 0.0}
                    )
                )
            )

        # Wait for hallucinations to come back and bundle them into a result list.
        processed_results = []
        for future in futures:
            hallucinations_output = await future
            processed_results.append(hallucinations_output.results[0])
        return ChatCompletionResults(results=processed_results)



# SPDX-License-Identifier: Apache-2.0

"""
I/O processor for the Granite query rewrite intrinsic.
"""

# Standard
import json

# Third Party
import pydantic

# Local
from granite_io.io.base import (
    ModelDirectInputOutputProcessorWithGenerate,
)
from granite_io.io.granite_3_2.input_processors.granite_3_2_input_processor import (
    Granite3Point2InputProcessor,
    Granite3Point2Inputs,
)
from granite_io.types import (
    ChatCompletionInputs,
    ChatCompletionResult,
    ChatCompletionResults,
    GenerateInputs,
    GenerateResults,
)

INSTRUCTION_TEXT = (
    "Reword the final utterance from the USER into a single utterance that doesn't "
    "need the prior conversation history to understand the user's intent. If the final "
    "utterance is a clear and standalone question, please DO NOT attempt to rewrite "
    "it, rather output the last user utterance as is. "
)
JSON = 'Your output format should be in JSON: { "rewritten_question": <REWRITE> }'
REWRITE_PROMPT = (
    "<|start_of_role|>rewrite: " + INSTRUCTION_TEXT + JSON + "<|end_of_role|>"
)


class QueryRewriteRawOutput(pydantic.BaseModel):
    rewritten_question: str


RAW_OUTPUT_JSON_SCHEMA = QueryRewriteRawOutput.model_json_schema()


class QueryRewriteIOProcessor(ModelDirectInputOutputProcessorWithGenerate):
    """
    I/O processor for the query rewrite intrinsic, also known as the [LoRA Adapter for 
    Query Rewrite](https://huggingface.co/ibm-granite/
    granite-3.2-8b-lora-rag-query-rewrite)

    Takes as input a chat completion and returns a completion with a rewrite of the
    most recent user turn (last in the conversation).

    Example raw input:
    ```
    <|start_of_role|>user<|end_of_role|>Tim Cook is the CEO of Apple Inc.<|end_of_text|>
    <|start_of_role|>assistant<|end_of_role|>Yes, Tim Cook is the Chief Executive \
Officer of Apple Inc.<|end_of_text|>
    <|start_of_role|>user<|end_of_role|>and for Microsoft?<|end_of_text|>
    ```
    
    Example of corresponding raw output:
    ```
    { "rewritten_question": "Who is the CEO of Microsoft" }
    ```

    Output string with the rewrite of the last user turn:
        'Who is the CEO of Microsoft'
    """

    def __init__(self, backend):
        super().__init__(backend=backend)

        # Input processor for the base model, which does most of the input formatting.
        self.base_input_processor = Granite3Point2InputProcessor()

    def inputs_to_generate_inputs(
        self, inputs: ChatCompletionInputs, add_generation_prompt: bool = True
    ) -> GenerateInputs:
        # Validate the input and convert to Granite input
        inputs = Granite3Point2Inputs.model_validate(inputs.model_dump())

        # Check for the invariants that the model expects its input to satisfy
        if not inputs.messages[-1].role == "user":
            raise ValueError("Last message is not a user message")

        # The beginning of the prompt doesn't change relative to base Granite 3.2
        prompt = self.base_input_processor.transform(inputs, False)

        # To invoke the model, we add the rewrite prompt to the prompt prefix:
        if add_generation_prompt:
            prompt = prompt + REWRITE_PROMPT
        result = inputs.generate_inputs.model_copy(
            update={
                "prompt": prompt,
                "max_tokens": 256,
                # TODO Enable constrained decoding on vLLM backends
                "extra_body": {
                    "guided_json": RAW_OUTPUT_JSON_SCHEMA,
                },
            }
        )
        return result

    def output_to_result(
        self,
        output: GenerateResults,
        inputs: ChatCompletionInputs | None = None,  # pylint: disable=unused-argument
    ) -> ChatCompletionResults:
        results = []
        for raw_result in output.results:
            json_result = raw_result.completion_string
            # print(f"{json_result=}")
            try:
                rewrite = json.loads(json_result)["rewritten_question"]
            except Exception as e:  # pylint: disable=broad-exception-caught
                # Cannot parse the JSON. Pass through the unparsed raw output for now
                # instead of raising an exception from the I/O processor.
                print(f"\nException: {e}\n")
                rewrite = json_result

            # Change content but retain other properties of the message.
            rewritten_last_message = inputs.messages[-1].model_copy(
                update={"content": rewrite}
            )
            results.append(ChatCompletionResult(next_message=rewritten_last_message))

        return ChatCompletionResults(results=results)



# SPDX-License-Identifier: Apache-2.0

"""
Tests for the majority voting I/O processor
"""

# Standard
import re

# Third Party
import pytest

# Local
from granite_io import (
    get_input_processor,
    get_output_processor,
    make_backend,
    make_new_io_processor,
)
from granite_io.backend import Backend
from granite_io.types import ChatCompletionInputs, ChatCompletionResults

ollama_model_name = "granite3.2:8b"
general_model_name = "Granite 3.2"


# All the different chat completion requests that are tested in this file, serialized as
# JSON strings. Represented as a dictionary instead of a list so that pytest output will
# show the short key instead of the long value when referencing a single run of a test
INPUT_JSON_STRS = {
    "simple": """
{
    "messages":
    [
        {"role": "user", "content": "Hello, how are you?"},
        {"role": "assistant", "content": "I'm doing great. How can I help you today?"},
        {"role": "user", "content": "I'd like to show off how chat templating works!"}
    ]
}
""",
    "thinking_tag": """
{
    "messages":
    [
        {"role": "user", "content": "How much wood could a wood chuck chuck?"}
    ],
    "thinking": true
}
""",
    "stop_strings": """
{
    "messages":
        [
            {"role": "user", "content": "How much wood could a wood chuck chuck?"}
        ],
    "generate_inputs":
        {
            "stop": "woodchuck"
        }
}
""",
}


@pytest.fixture(scope="session", params=INPUT_JSON_STRS)
def input_json_str(request: pytest.FixtureRequest) -> str:
    """Pytest fixture that allows us to run a given test case repeatedly with multiple
    different chat completion requests."""
    return INPUT_JSON_STRS[request.param]


@pytest.mark.vcr
def test_make_io_processor(backend_x: Backend, input_json_str: str):  # pylint: disable=redefined-outer-name
    inputs = ChatCompletionInputs.model_validate_json(input_json_str)
    io_processor = make_new_io_processor(
        input_processor=get_input_processor(general_model_name),
        output_processor=get_output_processor(general_model_name),
        backend=backend_x,
    )
    outputs: ChatCompletionResults = io_processor.create_chat_completion(inputs)

    assert isinstance(outputs, ChatCompletionResults)
    assert len(outputs.results) == 1

    content = outputs.results[0].next_message.content
    assert content  # Make sure we don't get empty result

    # Test for stop reason
    if inputs.generate_inputs and inputs.generate_inputs.stop:
        stop = inputs.generate_inputs.stop
        # Note: currently transformers includes the stop tokens,
        # but OpenAI does not. So, either endswith or not-in.
        assert stop not in content
        assert outputs.results[0].next_message.stop_reason == "stop"

    # TODO: Verify outputs in greater detail


def test_make_io_processor_no_input_processor():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Attempted to create IO Processor without setting an Input Processor"
        ),
    ):
        _ = make_new_io_processor(
            input_processor=None,
            output_processor=get_output_processor(general_model_name),
            backend=make_backend("openai", {"model_name": ollama_model_name}),
        )


def test_make_io_processor_no_output_processor():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Attempted to create IO Processor without setting an Output Processor"
        ),
    ):
        _ = make_new_io_processor(
            input_processor=get_input_processor(general_model_name),
            output_processor=None,
            backend=make_backend("openai", {"model_name": ollama_model_name}),
        )
